# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Utilities for comparison of content between changesets."""
import attr
from functools import cached_property
import re
import subprocess

from mercurial import (
    copies,
    diffutil,
    error,
    match as matchmod,
    patch as patchmod,
)

from .file_context import git_perms
from .git import (
    FILECTX_FLAGS_TO_GIT_MODE_BYTES,
    NULL_BLOB_OID,
    OBJECT_MODE_DOES_NOT_EXIST,
)
from .oid import (
    blob_oid,
    ctx_blob_oid,
)

from .stub.diff_pb2 import (
    ChangedPaths,
    DiffStats,
)

GIT_PATCH_ID_TIMEOUT_SECONDS = 10
"""Time given to `git patch-id` to finish after we've streamed the content.

Given that `git patch-id` is a C program (verified assertion), we expect it
to be much faster than HGitaly feeding it. A timeout of 10 seconds after
we're done streaming the patch content is therefore enormous.
"""

Status_Type_Map = dict(
    added=ChangedPaths.Status.ADDED,
    modified=ChangedPaths.Status.MODIFIED,
    removed=ChangedPaths.Status.DELETED,
    # Note: Mercurial includes TYPE_CHANGE
    # (symlink, regular file, submodule...etc) in MODIFIED status
)
"""Mapping status object attributes to ChangedPaths enum."""

COPIED = ChangedPaths.Status.COPIED
DIFF_HUNKS_START_RX = re.compile(rb'^(--- )|^(Binary file)')
"""To match the header line right before hunks start getting dumped."""


def changed_paths(repo, from_ctx, to_ctx, base_path):
    if base_path is None:
        matcher = None
        path_trim_at = 0
    else:
        # scmutil's match is more geared towards the CLI
        # `hg log` etc and its include patterns would
        # force us to convert everything to absolute paths
        # since $CWD is not the repo root.
        # It is much simpler to use the lower-level match module.
        matcher = matchmod.match(
            root=repo.root,
            # cwd is required, yet should not be relevant
            # in this case.
            cwd=repo.root,
            patterns=[b'path:' + base_path])
        path_trim_at = len(base_path) + 1

    copied = list(copy_changed_paths(
        from_ctx,
        to_ctx,
        copies.pathcopies(from_ctx, to_ctx, match=matcher),
        trim_at=path_trim_at))
    copied_paths = set(cp.path for cp in copied)

    status = from_ctx.status(to_ctx, match=matcher)
    for path in status_changed_paths(from_ctx, to_ctx, status,
                                     trim_at=path_trim_at):
        if path.path not in copied_paths:
            yield path

    yield from iter(copied)


def status_changed_paths(from_ctx, to_ctx, status, trim_at=0):
    """Return ChangedPaths from Mercurial status object"""
    for stype in ['added', 'modified', 'removed']:
        for path in status.__getattribute__(stype):
            if stype == 'added':
                old_mode = OBJECT_MODE_DOES_NOT_EXIST
                old_blob_id = NULL_BLOB_OID
            else:
                old_mode = git_perms(from_ctx.filectx(path))
                old_blob_id = ctx_blob_oid(from_ctx, path)

            if stype == 'removed':
                new_mode = OBJECT_MODE_DOES_NOT_EXIST
                new_blob_id = NULL_BLOB_OID
            else:
                new_mode = git_perms(to_ctx.filectx(path))
                new_blob_id = ctx_blob_oid(to_ctx, path)

            yield ChangedPaths(
                path=path[trim_at:],
                old_mode=old_mode,
                new_mode=new_mode,
                old_blob_id=old_blob_id,
                new_blob_id=new_blob_id,
                status=Status_Type_Map[stype]
            )


def copy_changed_paths(from_ctx, to_ctx, path_copies, trim_at=0):
    """Return ChangedPaths for the given paths, relative to base_path.

    Given that Gitaly currently (gitaly@c54d613d0) does not pass
    `--find-copies-harder` to `git diff-tree`, we cannot be sure of
    what is expected. That being said, `git diff-tree` gives the permission
    at source path as `old_mode`, so we're doing the same.
    """
    for target, source in path_copies.items():
        yield ChangedPaths(path=target[trim_at:],
                           status=COPIED,
                           old_mode=git_perms(to_ctx.filectx(source)),
                           new_mode=git_perms(to_ctx.filectx(target)),
                           )


def chunk_stats(chunks, from_ctx, to_ctx):
    """Yield the DiffStats messages from the given diff chunks.

    Changectx params are there for uniformity and not really needed as
    of this writing
    """
    for hg_chunk in patchmod.parsepatch(chunks):
        chunk = Chunk(hg_chunk, from_ctx, to_ctx)

        old_path, path = chunk.from_to_file_paths
        if old_path == path:
            old_path = b''
        adds, dels = chunk.additions_deletions()
        yield DiffStats(
            path=path,
            old_path=old_path,
            additions=adds,
            deletions=dels,
        )


def diff_opts(repo, git=True):
    opts = {b'git': git}
    return diffutil.difffeatureopts(repo.ui, opts=opts, git=git)


def _get_filectx(changeset, path):
    try:
        return changeset[path]
    except error.ManifestLookupError:
        return None


@attr.define
class Chunk:
    """Wrap a Mercurial chunk with extracted information.

    This notably avoids repeated lookups in manifests.

    Perfomance notes: we assume that looking up in the manifest is the
    expensive task, not building :class:`filectx` instances, hence we
    do not make methods to check if a file is in one of the changesets,
    just use, e.g.,  :meth:`from_filectx` and check for ``None``.
    """

    hg_chunk = attr.ib()
    from_ctx = attr.ib()
    to_ctx = attr.ib()

    @cached_property
    def from_to_file_paths(self):
        """Return a tuple of (old, new) file path for a diff chunk header

        Takes case of renames into account
        """
        header = self.hg_chunk
        fname = header.filename()
        from_path, to_path = fname, fname
        if len(header.files()) > 1:
            # file is renamed
            from_path, to_path = header.files()
        return from_path, to_path

    @cached_property
    def from_filectx(self):
        return _get_filectx(self.from_ctx, self.from_file_path)

    @cached_property
    def to_filectx(self):
        return _get_filectx(self.to_ctx, self.to_file_path)

    @property
    def from_file_path(self):
        return self.from_to_file_paths[0]

    @property
    def to_file_path(self):
        return self.from_to_file_paths[1]

    def from_to_blob_oids(self):
        from_bid = to_bid = NULL_BLOB_OID
        from_path, to_path = self.from_to_file_paths

        if self.from_filectx is not None:
            cid = self.from_ctx.hex().decode('ascii')
            from_bid = blob_oid(None, cid, from_path)
        if self.to_filectx is not None:
            cid = self.to_ctx.hex().decode('ascii')
            to_bid = blob_oid(None, cid, to_path)
        return from_bid, to_bid

    def from_to_file_mode(self):
        from_path, to_path = self.from_to_file_paths
        from_mode, to_mode = b'0', b'0'
        from_filectx = self.from_filectx
        if from_filectx is not None:
            from_mode = FILECTX_FLAGS_TO_GIT_MODE_BYTES[from_filectx.flags()]
        to_filectx = self.to_filectx
        if to_filectx is not None:
            to_mode = FILECTX_FLAGS_TO_GIT_MODE_BYTES[to_filectx.flags()]
        return from_mode, to_mode

    def additions_deletions(self):
        """Return the pair (addition, deletions) for the chunk."""
        adds, dels = 0, 0
        for hunk in self.hg_chunk.hunks:
            add_count, del_count = hunk.countchanges(hunk.hunk)
            adds += add_count
            dels += del_count
        return adds, dels

    def header_with_index_line(self):
        """Generate header with the index line and binary indication.

        The index line is the expected Git-style one, with file modes etc.
        The binary indication tells whether there should be a placeholder
        instead actual Git binary content section. Callers can use it to
        generate the appropriate placeholder for their needs.
        """
        fname = self.hg_chunk.filename()
        old_bid, new_bid = self.from_to_blob_oids()
        indexline = ('index %s..%s' % (old_bid, new_bid)).encode('ascii')

        # Note: <mode> is required only when it didn't change between
        # the two changesets, otherwise it has a separate line
        if self.to_filectx is not None and self.to_filectx.path() == fname:
            oldmode, mode = self.from_to_file_mode()
            if mode == oldmode:
                indexline += b' ' + mode
        indexline += b'\n'
        headerlines = self.hg_chunk.header

        binary = False
        for index, line in enumerate(headerlines[:]):
            m = DIFF_HUNKS_START_RX.match(line)
            if m is None:
                continue

            binary = not bool(m.group(1))
            headerlines.insert(index, indexline)
            break
        return b''.join(headerlines), binary


def write_diff_to_file(fobj, changeset_from, changeset_to, dump_binary=True):
    """Compute diff and stream it to a file-like object

    The diff includes the expected "index line" as Git does: change of OID
    and of permissions for each file.

    :param fobj: a file-like object
    :param dump_binary: if ``True``, the Git binary content is dumped.
      Otherwise a placeholder is inserted, made of the file node ids. This
      matches Gitaly's behaviour in ``GetPatchId`` implementation (quoting
      internal/gitaly/service/diff/patch_id.go as of v16.6)::

        // git-patch-id(1) will ignore binary diffs, and computing binary
        // diffs would be expensive anyway for large blobs. This means that
        // we must instead use the pre- and post-image blob IDs that
        // git-diff(1) prints for binary diffs as input to git-patch-id(1),
        // but unfortunately this is only honored in Git v2.39.0 and newer.
        // We have no other choice than to accept this though, so we instead
        // just ask git-diff(1) to print the full blob IDs for the pre- and
        // post-image blobs instead of abbreviated ones so that we can avoid
        // any kind of potential prefix collisions.
        git.Flag{Name: "--full-index"},
    """
    repo = changeset_from.repo()
    # hg diff --git --no-binary does not include the index line
    # hg diff --git does include the index line but also dumps binary
    #   content, which is uselessly expensive in some cases (GetPatchID)
    # TODO generally avoid actual binary content in DiffService
    #   when Gitaly does the same.
    low_level_diff_opts = diff_opts(repo)
    low_level_diff_opts.nobinary = not dump_binary

    hg_chunks = changeset_to.diff(changeset_from, opts=low_level_diff_opts)
    for hg_chunk in patchmod.parsepatch(hg_chunks):
        chunk = Chunk(hg_chunk, changeset_from, changeset_to)
        from_path, to_path = chunk.from_to_file_paths
        header, binary_placeholder = chunk.header_with_index_line()
        fobj.write(header)

        if binary_placeholder:
            filename = hg_chunk.filename()
            fobj.write(b'--- a/%s\n' % filename)
            fobj.write(b'+++ b/%s\n' % filename)
            fobj.write(b'@@ -1 +1 @@\n')
            from_bid = to_bid = NULL_BLOB_OID.encode()
            if chunk.from_filectx is not None:
                from_bid = chunk.from_filectx.hex()
            if chunk.to_filectx is not None:
                to_bid = chunk.to_filectx.hex()
            fobj.write(b'-%s\n' % from_bid)
            fobj.write(b'+%s\n' % to_bid)
        else:
            for hunk in hg_chunk.hunks:
                hunk.write(fobj)


def run_git_patch_id(git_path, writer):
    """Call `git patch-id` in a subprocess

    :param writer: a callable writing the diff content to a file-like object
    """
    git = subprocess.Popen((git_path, b'patch-id'),
                           stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
    writer(git.stdin)

    try:
        # this also takes care of flushing/closing stdin
        out, err = git.communicate(timeout=GIT_PATCH_ID_TIMEOUT_SECONDS)
    except subprocess.TimeoutExpired:
        # Quoting https://docs.python.org/3/library/subprocess.html:
        #   The child process is not killed if the timeout expires,
        #   so in order to cleanup properly a well-behaved application
        #   should kill the child process and finish communication
        git.kill()
        git.communicate()
        raise

    if git.returncode != 0:
        raise RuntimeError("git-patch-id returned code %d, stderr=%r" % (
            git.returncode, err))

    return out.strip().decode('ascii')


def git_patch_id(git_path, changeset_from, changeset_to):
    return run_git_patch_id(
        git_path,
        lambda stdin: write_diff_to_file(stdin, changeset_from, changeset_to,
                                         dump_binary=False)
    )



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest
from grpc import StatusCode

from ..license_detector import detect_license
from ..testing.context import (
    FakeContextAborter as FakeContext,
    FakeContextAbort,
)


def test_detector_subprocesss_errors(tmpdir):
    bin_path = tmpdir / 'detector'
    context = FakeContext()
    with pytest.raises(FakeContextAbort):
        detect_license(bin_path, tmpdir / 'tree', context)
    assert context.code() == StatusCode.INTERNAL
    assert "not found" in context.details()

    bin_path.write_binary(b'#!/bin/sh\n\n echo "oops" >&2; exit 3\n')
    with pytest.raises(FakeContextAbort):
        detect_license(bin_path, tmpdir / 'tree', context)
    assert "permission denied" in context.details()

    bin_path.chmod(0o755)

    with pytest.raises(FakeContextAbort):
        detect_license(bin_path, tmpdir / 'tree', context)
    assert context.code() == StatusCode.INTERNAL
    assert "status code=3" in context.details()
    assert "oops" in context.details()

    bin_path.write_binary(b'#!/bin/sh\n\n echo \'[{"error": "testerr"}]\'\n')
    with pytest.raises(FakeContextAbort):
        detect_license(bin_path, tmpdir / 'tree', context)
    assert context.code() == StatusCode.INTERNAL
    assert 'testerr' in context.details()



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from grpc import StatusCode
import logging
import time

from mercurial.merge import merge
from mercurial.mergestate import mergestate
from mercurial.node import nullhex as NULL_NODE_HEX
from mercurial.phases import (
    public as PUBLIC,
)
from mercurial import util as hgutil

from heptapod.gitlab.branch import (
    NAMED_BRANCH_PREFIX,
    gitlab_branch_ref,
)

from .. import message
from ..branch import (
    gitlab_branch_head
)
from ..changelog import (
    ancestor,
    merge_content,
)
from ..errors import (
    operation_error_treatment,
    structured_abort,
)
from ..logging import LoggerAdapter
from ..revision import (
    changeset_by_commit_id_abort,
    gitlab_revision_changeset,
    validate_oid,
)
from ..stub.errors_pb2 import (
    MergeConflictError,
    ReferenceUpdateError,
)
from ..stub.mercurial_operations_pb2 import (
    CensorRequest,
    CensorResponse,
    MercurialPermissions,
    MergeAnalysisRequest,
    MergeAnalysisResponse,
    MergeBranchError,
    MergeBranchRequest,
    MergeBranchResponse,
    PreCheckUpdateError,
    PublishChangesetError,
    PublishChangesetRequest,
    PublishChangesetResponse,
)
from ..stub.mercurial_operations_pb2_grpc import (
    MercurialOperationsServiceServicer,
)
from ..stub.operations_pb2 import (
    OperationBranchUpdate,
)
from ..stub.shared_pb2 import (
    User,
)
from ..servicer import HGitalyServicer

base_logger = logging.getLogger(__name__)

MERCURIAL_VERSION = hgutil.versiontuple()
MERGE_CONFLICTS_LABELS = [b'working copy', b'merge rev', b'common ancestor']


class ConflictError(RuntimeError):
    """Raised by merge helpers and catched by service method"""


class MercurialOperationsServicer(MercurialOperationsServiceServicer,
                                  HGitalyServicer):
    """MercurialOperationService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """
    def MergeAnalysis(self,
                      request: MergeAnalysisRequest,
                      context) -> MergeAnalysisResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)
        source_cs = gitlab_revision_changeset(repo, request.source_revision)
        if source_cs is None:
            context.abort(
                StatusCode.INVALID_ARGUMENT,
                "Source revision %r not found" % request.source_revision)
        target_cs = gitlab_revision_changeset(repo, request.target_revision)
        if target_cs is None:
            context.abort(
                StatusCode.INVALID_ARGUMENT,
                "Target revision %r not found" % request.target_revision)

        source_branch = source_cs.branch()
        target_branch = target_cs.branch()
        logger.info("Merge Analysis: source branch %r, target branch %r",
                    source_branch, target_branch)
        is_ff = (ancestor(source_cs, target_cs) == target_cs.rev()
                 and source_branch == target_branch)

        has_obsolete = has_unstable = False
        for cs in merge_content(source_cs, target_cs):
            if not has_obsolete:
                has_obsolete = cs.obsolete()
            if not has_unstable:
                has_unstable = cs.isunstable()
            if has_obsolete and has_unstable:
                break

        has_conflicts = False
        if not (is_ff
                or has_obsolete or has_unstable
                or request.skip_conflicts_check):
            with self.working_dir(gl_repo=request.repository,
                                  repo=repo,
                                  context=context,
                                  changeset=target_cs) as wd:
                has_conflicts = not wd_merge(wd, source_cs)

        res = MergeAnalysisResponse(
            is_fast_forward=is_ff,
            has_obsolete_changesets=has_obsolete,
            has_unstable_changesets=has_unstable,
            has_conflicts=has_conflicts,
            target_is_public=target_cs.phase() == PUBLIC,
            target_node_id=target_cs.hex().decode('ascii'),
            target_branch=target_cs.branch(),
            target_topic=target_cs.topic(),
            source_node_id=source_cs.hex().decode('ascii'),
            source_branch=source_cs.branch(),
            source_topic=source_cs.topic(),
        )
        logger.info("MergeAnalysis result %r", message.Logging(res))
        return res

    def PublishChangeset(self,
                         request: PublishChangesetRequest,
                         context) -> PublishChangesetResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context,
                              for_mutation_by=request.user)
        with_hg_git = not repo.ui.configbool(b'heptapod', b'no-git')

        gl_rev = request.gitlab_revision
        if not gl_rev:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty gitlab_revision")

        to_publish = gitlab_revision_changeset(repo, request.gitlab_revision)
        if to_publish is None:
            context.abort(
                StatusCode.INVALID_ARGUMENT,
                f"could not resolve revision {repr(request.gitlab_revision)}"
            )

        # TODO PermissionDenied, then
        if request.hg_perms < MercurialPermissions.PUBLISH:
            context.abort(StatusCode.FAILED_PRECONDITION,
                          "Publish permission not passed")

        logger.info("publishing %r, mirroring to Git=%r", to_publish,
                    with_hg_git)

        with operation_error_treatment(context, PublishChangesetError,
                                       logger=logger):
            self.repo_command(repo, context, 'phase',
                              public=True,
                              draft=False,
                              secret=False,
                              force=False,
                              rev=[str(to_publish.rev()).encode('ascii')])
        return PublishChangesetResponse()

    def Censor(self, request: CensorRequest, context) -> CensorResponse:
        # censor won't even change a hash, still let's follow the usual
        # procedure, in case there turned out to be a side effect
        repo = self.load_repo(request.repository, context,
                              for_mutation_by=request.user)
        node_id = request.changeset_node_id
        changeset = changeset_by_commit_id_abort(repo, node_id, context)
        if changeset is None:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          f"Changeset {node_id} not found")
        # it is essential that the censored file is not checked out.
        # normally, the repository should already be at the null revision,
        # but let's be really sure:
        self.repo_command(repo, context, 'update', NULL_NODE_HEX)

        if MERCURIAL_VERSION >= (6, 7):
            rev_arg = [changeset.rev()]  # hg>=6.7
        else:
            rev_arg = changeset.hex()  # hg<6.7

        self.repo_command(repo, context, 'censor',
                          b'/'.join((repo.root, request.file_path)),
                          rev=rev_arg,
                          tombstone=request.tombstone)
        return CensorResponse()

    def MergeBranch(self,
                    request: MergeBranchRequest,
                    context) -> MergeBranchResponse:
        logger = LoggerAdapter(base_logger, context)

        repo = self.load_repo(request.repository, context,
                              for_mutation_by=request.user)
        with_hg_git = not repo.ui.configbool(b'heptapod', b'no-git')

        target_branch = request.branch
        if not target_branch:
            context.abort(StatusCode.INVALID_ARGUMENT, "empty branch name")
        if not target_branch.startswith(NAMED_BRANCH_PREFIX):
            context.abort(StatusCode.FAILED_PRECONDITION,
                          "Heptapod merges are currently targeting "
                          "named branches only (no topics nor bookmarks)")

        if request.hg_perms < MercurialPermissions.PUBLISH:
            context.abort(StatusCode.PERMISSION_DENIED,
                          "Publish permission not passed")

        if not request.message:
            # would not be necessary if we turn out to fast-forward
            # but that cannot be known by the caller in advance
            context.abort(StatusCode.INVALID_ARGUMENT, "empty message")

        to_merge = changeset_by_commit_id_abort(
            repo, request.commit_id, context)
        if to_merge is None:
            context.abort(
                StatusCode.INTERNAL,
                f'checking for ancestry: invalid commit: "{request.commit_id}"'
            )

        old_id = request.expected_old_oid
        if old_id and not validate_oid(old_id):
            context.abort(StatusCode.INVALID_ARGUMENT,
                          f'cannot parse commit ID: "{old_id}"')

        current_head = gitlab_branch_head(repo, request.branch)
        # TODO mutualize as a check_obsolete_instability()
        # or check_publishable()
        for cs in merge_content(to_merge, current_head):
            if cs.obsolete():
                abort_pre_check_changeset(
                    context, cs,
                    cause=PreCheckUpdateError.OBSOLETE_CHANGESET,
                    error_cls=MergeBranchError
                )
            if cs.isunstable():
                abort_pre_check_changeset(
                    context, cs,
                    cause=PreCheckUpdateError.UNSTABLE_CHANGESET,
                    error_cls=MergeBranchError
                )

        current_head_id = current_head.hex().decode('ascii')
        if old_id and old_id != current_head_id:
            # We did not need to resolve before this, but now we do because
            # Gitaly has a specific error if resolution fails
            if changeset_by_commit_id_abort(repo, old_id, context) is None:
                context.abort(StatusCode.INVALID_ARGUMENT,
                              "cannot resolve expected old object ID: "
                              "reference not found")
            structured_abort(
                context,
                StatusCode.FAILED_PRECONDITION,
                "reference update: reference does not point to "
                "expected object",
                MergeBranchError(
                    reference_check=ReferenceUpdateError(
                        reference_name=gitlab_branch_ref(request.branch),
                        old_oid=old_id,
                        # Gitaly would set `new_oid` to the just created
                        # merge commit id. We cannot do that, let's put
                        # something potentially useful
                        new_oid=current_head_id,
                    )))

        if ancestor(to_merge, current_head) == current_head.rev():
            # fast-forwardable!
            if not request.semi_linear:
                logger.info("This is actually a fast-forward, all checks "
                            "passed, now publishing %r, mirroring to Git=%r",
                            to_merge, with_hg_git)
                self.publish(to_merge, context)
                return MergeBranchResponse(
                    branch_update=OperationBranchUpdate(
                        commit_id=to_merge.hex().decode('ascii'),
                    ))
        elif request.semi_linear:
            structured_abort(
                context, StatusCode.FAILED_PRECONDITION,
                "Not fast-forwardable",
                MergeBranchError(
                    pre_check=PreCheckUpdateError.NOT_FAST_FORWARD)
            )

        # actual merge
        with self.working_dir(gl_repo=request.repository,
                              repo=repo,
                              changeset=current_head,
                              context=context) as wd:
            # `allowunstable=no` protects us against all instabilities,
            # in particular against orphaning dependent topics.
            # TODO this setting should probably be set in all mutations
            wd.repo.ui.setconfig(b'experimental.evolution', b'allowunstable',
                                 False)
            # allowing so-called oedipian merge (for semi-linear use case)
            wd.repo.ui.setconfig(b'experimental', b'topic.linear-merge',
                                 b'allow-from-bare-branch')
            if not request.HasField('timestamp'):
                unix_ts = int(time.time())
            else:
                unix_ts = request.timestamp.seconds
            # TODO we should do everything in a single transaction
            # (even ideally including the checks) but this is already so much
            # better than the several `hg` processes that this method replaces
            # that we can postpone that to a later move.

            # Passing as revision number will reacquire the changeset, which
            # is really needed as we are now in a different repo (share).
            try:
                # TODO test we don't wait on some impossible user input
                with operation_error_treatment(context, MergeBranchError,
                                               logger=logger):
                    new_changeset = self.merge(working_dir=wd,
                                               source_changeset=to_merge,
                                               user=request.user,
                                               message=request.message,
                                               unix_timestamp=unix_ts,
                                               context=context)
            except ConflictError as exc:
                unresolved = exc.args[0]
                structured_abort(
                    context,
                    StatusCode.FAILED_PRECONDITION,
                    "merging commits: merge: there are conflicting files",
                    MergeBranchError(
                        conflict=MergeConflictError(
                            conflicting_commit_ids=[
                                cs.hex().decode('ascii')
                                for cs in (current_head, to_merge)
                            ],
                            conflicting_files=unresolved
                        )
                    )
                )

            self.publish(new_changeset, context)
            return MergeBranchResponse(
                branch_update=OperationBranchUpdate(
                    commit_id=new_changeset.hex().decode('ascii'),
                ))

    def merge(self, working_dir, context, source_changeset,
              user: User, message, unix_timestamp):
        """Merge changeset with given revno in working repository.

        The working repositoy is assumed to be already updated in a
        meaningful way.
        """
        working_repo = working_dir.repo
        if not wd_merge(working_dir, source_changeset):
            unresolved = list(mergestate.read(working_repo).unresolved())
            self.repo_command(working_repo, context, 'merge', abort=True)
            raise ConflictError(unresolved)
        # TODO user timezone (symbolic or UTC offset?)
        self.repo_command(working_repo, context, 'commit',
                          user=b'%s <%s>' % (user.name, user.email),
                          date=b'%d 0' % unix_timestamp,
                          message=message)
        return working_repo[b'.']


def wd_merge(working_dir, source_cs):
    """Merge source_cs in the given a working directory (repo share).

    :source_cs: a :class:`changectx`, usually not tied to ``working_dir`` but
      to its share source or a share sibling.
    :return: whether it suceeded
    """
    # re-evalutate the changectx in the context of working_dir,
    # as `merge()` will read the repo from it
    repo = working_dir.repo
    source_for_wd = repo[source_cs.rev()]

    overrides = {(b'ui', b'forcemerge'): b'internal:merge3',
                 (b'ui', b'interactive'): b'off'}
    with repo.ui.configoverride(overrides, b'merge'):
        # not sure labels are really necessary, but it is
        # possible that the merge tools require them.
        stats = merge(source_for_wd,
                      labels=MERGE_CONFLICTS_LABELS)
        return not stats.unresolvedcount


def abort_pre_check_changeset(context, changeset, cause, error_cls):
    structured_abort(
        context,
        StatusCode.FAILED_PRECONDITION,
        changeset.hex().decode('ascii'),
        error_cls(pre_check=cause)
    )



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from contextlib import contextmanager
import gc
import logging
import os
from pathlib import Path
import time
import weakref

from grpc import StatusCode
from mercurial import (
    cmdutil,
    commands,
    error,
    hg,
    ui as uimod,
)
from mercurial.repoview import _filteredrepotypes

from .errors import ServiceError
from .logging import LoggerAdapter
from .workdir import working_directory
from .stub.shared_pb2 import (
    Repository,
)

GARBAGE_COLLECTING_RATE_GEN2 = 250
GARBAGE_COLLECTING_RATE_GEN1 = 20
HG_GIT_MIRRORING_MD_KEY = 'x-heptapod-hg-git-mirroring'
NATIVE_PROJECT_MD_KEY = 'x-heptapod-hg-native'
# TODO check if there is a GitLab MD for this, it should:
SKIP_HOOKS_MD_KEY = 'x-heptapod-skip-gl-hooks'
ACCEPT_MR_IID_KEY = 'x-heptapod-accept-mr-iid'
# TODO expose it in heptpod.wsgi and import it here
PY_HEPTAPOD_SKIP_HOOKS = b'HEPTAPOD_SKIP_ALL_GITLAB_HOOKS'

logger = logging.getLogger(__name__)

USER_DATA_TO_ENVIRON = {
}
PROJECT_DATA_TO_ENVIRON = {
}

CONFIG_TO_FORWARD_IN_WORKING_DIRS = ((b'heptapod', b'no-git'),
                                     (b'heptapod', b'native'))


def clear_repo_class(repo_class):
    _filteredrepotypes.pop(repo_class, None)


repos_counter = 0


def gc_collect(generation):
    logger.info("A total of %d repository objects have been "
                "instantiated in this thread since startup. "
                "Garbage collecting (generation %d)",
                repos_counter, generation)
    start = time.time()
    gc_result = gc.collect(generation=generation)
    logger.info("Garbage collection (generation %d) "
                "done in %f seconds "
                "(%d unreachable objects). Current GC stats: %r",
                generation, time.time() - start, gc_result, gc.get_stats())


def normalize_rpath(gl_repo: Repository):
    rpath = gl_repo.relative_path
    if rpath.endswith('.git'):
        rpath = rpath[:-4] + '.hg'
    return rpath


class RepositoryCreationError(ServiceError):
    """Specific exception for creation problems."""


class HGitalyServicer:
    """Common features of all HGitaly services.

    Attributes:

    - :attr:`storages`: a :class:`dict` mapping storage names to corresponding
      root directory absolute paths, which are given as bytes since we'll have
      to convert paths to bytes anyway, which is the only promise a filesystem
      can make, and what Mercurial expects.
    - :attr:`ui`: base :class:`mercurial.ui.ui` instance from which repository
      uis are derived. In particular, it bears the common configuration.
    """

    STATUS_CODE_STORAGE_NOT_FOUND = StatusCode.NOT_FOUND
    STATUS_CODE_REPO_NOT_FOUND = StatusCode.NOT_FOUND
    STATUS_CODE_MISSING_REPO_ARG = StatusCode.INVALID_ARGUMENT

    def __init__(self, storages):
        self.storages = storages
        self.init_ui()

    def init_ui(self):
        """Prepare the base ``ui`` instance from which all will derive.

        See also :meth:`hgweb.__init__`
        """

        ui = self.ui = uimod.ui.load()
        # prevents `ui.interactive()` to crash (see heptapod#717)
        ui.setconfig(b'ui', b'nontty', b'true', b'hgitaly')

        # progress bars in stdout (log file at best) would be rather useless
        ui.setconfig(b'progress', b'disable', b'true', b'hgitaly')

        # other settings done in hgweb.__init__():
        #
        # - forcing file pattern resolution to be relative to root would be
        #   nice, but perhaps need more adaptation, and would have to be
        #   done in load_repo()
        # - `signal-safe-locks=no` worth being considered, but this not being
        #   WSGI, we control the server and its signal handling (see hgweb's
        #   comment)
        # - `report_unstrusted=off`: if some perms are unaligned, reporting
        #   the untrust could be the only lead for an obscure behaviour
        #   (typically ignoring some settings that can be critical to
        #   operation)

    def load_repo(self, repository: Repository, context,
                  for_mutation_by=None,
                  ):
        """Load the repository from storage name and relative path

        :param repository: Repository Gitaly Message, encoding storage name
            and relative path
        :param context: gRPC context (used in error raising)
        :raises: ``KeyError('storage', storage_name)`` if storage is not found.

        Error treatment: the caller doesn't have to do anything specific,
        the status code and the details are already set in context, and these
        are automatically propagated to the client (see corresponding test
        in `test_servicer.py`). For specific error treatment, use
        :meth:`load_repo_inner` and catch the exceptions it raises.
        """
        try:
            repo = self.load_repo_inner(repository, context)
        except KeyError as exc:
            self.handle_key_error(context, exc.args)
        except ValueError as exc:
            self.handle_value_error(context, exc.args)

        if for_mutation_by is not None:
            if not for_mutation_by.gl_id:
                context.abort(StatusCode.INVALID_ARGUMENT, "empty User")
            self.add_mutation_metadata(grpc_repo=repository,
                                       user=for_mutation_by,
                                       ui=repo.ui,
                                       context=context)

        return repo

    def add_mutation_metadata(self, grpc_repo, user, context, ui):
        """Store metadata needed for mutation in repo environ.

        This is roughly the same job as heptapod.wsgi does.

        If we ever cache the repository objects, we will need to rerun this
        and/or reclone ``ui`` from a base config.
        """
        mirroring = native = skip_gl_hooks = False
        accept_mr_iid = None
        for md in context.invocation_metadata():
            if md.key == HG_GIT_MIRRORING_MD_KEY:
                mirroring = md.value.strip().lower() == 'true'
            elif md.key == NATIVE_PROJECT_MD_KEY:
                native = md.value.strip().lower() == 'true'
            elif md.key == SKIP_HOOKS_MD_KEY:
                skip_gl_hooks = md.value.strip().lower() == 'true'
            elif md.key == ACCEPT_MR_IID_KEY:
                accept_mr_iid = md.value.strip()

        mirroring = mirroring or not native

        ui.setconfig(b'heptapod', b'native', native)
        ui.setconfig(b'heptapod', b'no-git', not mirroring)

        environ = ui.environ
        if skip_gl_hooks:
            environ[PY_HEPTAPOD_SKIP_HOOKS] = b'yes'

        environ[b'HEPTAPOD_USERINFO_GL_ID'] = user.gl_id.encode('ascii')
        environ[b'HEPTAPOD_USERINFO_USERNAME'] = (
            user.gl_username.encode('utf-8'))
        environ[b'HEPTAPOD_USERINFO_NAME'] = user.name
        environ[b'HEPTAPOD_USERINFO_EMAIL'] = user.email
        environ[b'GL_PROTOCOL'] = b'web'  # same hardcoding as in Gitaly
        environ[b'GL_REPOSITORY'] = grpc_repo.gl_repository.encode('utf-8')
        # TODO project but apparently not used in py-heptapod hooks but still
        # may be useful for user-defined hooks
        if accept_mr_iid is not None:
            environ[b'HEPTAPOD_ACCEPT_MR_IID'] = accept_mr_iid.encode('ascii')

    def handle_value_error(self, context, exc_args):
        context.abort(self.STATUS_CODE_MISSING_REPO_ARG, exc_args[0])

    def handle_key_error(self, context, exc_args):
        ktype = exc_args[0]
        if ktype == 'storage':
            context.abort(self.STATUS_CODE_STORAGE_NOT_FOUND,
                          "No storage named %r" % exc_args[1])
        elif ktype == 'repo':
            context.abort(self.STATUS_CODE_REPO_NOT_FOUND,
                          exc_args[1])

    def create_and_load_repo(self, repository: Repository, context):
        """Create the repository and load it.

        :raises RepositoryCreationError:
        """
        self.hg_init_repository(repository, context)
        return self.load_repo_inner(repository, context)

    def load_or_create_repo(self, repository: Repository, context):
        """Load the repository, creating it if needed.

        :returns: a boolean being `True` iff creation occurred and the
        repository itself.
        """
        try:
            return False, self.load_repo_inner(repository, context)
        except KeyError as exc:
            if exc.args[0] == 'repo':
                return True, self.create_and_load_repo(repository, context)
            else:
                self.handle_key_error(context, exc.args)
        except ValueError as exc:
            self.handle_value_error(context, exc.args)

    def load_repo_inner(self, repository: Repository, context):
        """Load the repository from storage name and relative path

        :param repository: Repository Gitaly Message, encoding storage name
            and relative path
        :param context: gRPC context (used in error raising)
        :raises:
          - ``KeyError('storage', storage_name)`` if storage is not found
          - ``KeyError('repo', path, details)`` if repo not found or
            cannot be loaded.
        """
        global repos_counter
        if repos_counter % GARBAGE_COLLECTING_RATE_GEN2 == 0:
            gc_collect(2)
        elif repos_counter % GARBAGE_COLLECTING_RATE_GEN1 == 0:
            gc_collect(1)

        repos_counter += 1

        # shamelessly taken from heptapod.wsgi for the Hgitaly bootstrap
        # note that Gitaly Repository has more than just a relative path,
        # we'll have to decide what we make of the extra information
        repo_path = self.repo_disk_path(repository, context)
        logger.debug("loading repo at %r", repo_path)

        try:
            repo = hg.repository(self.ui, repo_path)
        except error.RepoError as exc:
            raise KeyError('repo', repo_path, repr(exc.args))

        weakref.finalize(repo, clear_repo_class, repo.unfiltered().__class__)
        srcrepo = hg.sharedreposource(repo)
        if srcrepo is not None:
            weakref.finalize(srcrepo, clear_repo_class,
                             srcrepo.unfiltered().__class__)

        return repo

    def storage_root_dir(self, storage_name, context):
        """Return the storage directory.

        If the storage is unknown, this raises
        ``KeyError('storage', storage_name)``
        """
        if not storage_name:
            # this is the best detection of a missing `repository` field
            # in request, without having the request object itself
            raise ValueError('repository not set')

        root_dir = self.storages.get(storage_name)
        if root_dir is None:
            raise KeyError('storage', storage_name)
        return root_dir

    def repo_disk_path(self, repository: Repository, context):
        rpath = normalize_rpath(repository)
        root_dir = self.storage_root_dir(repository.storage_name, context)

        # GitLab filesystem paths are always ASCII
        repo_path = os.path.join(root_dir, rpath.encode('ascii'))
        return repo_path

    def temp_dir(self, storage_name, context, ensure=True):
        """Return the path to temporary directory for the given storage

        Similar to what Gitaly uses, with a dedicated path in order
        to be really sure not to overwrite anything. The important feature
        is that the temporary directory is under the root directory of
        the storage, hence on the same file system (atomic renames of
        other files from the storage, etc.)

        :param bool ensure: if ``True``, the temporary directory is created
           if it does not exist yet.
        """
        try:
            return self.temp_dir_inner(storage_name, context, ensure=ensure)
        except KeyError as exc:
            self.handle_key_error(context, exc.args)
        except ValueError as exc:
            self.handle_value_error(context, exc.args)
        except OSError as exc:
            context.abort(StatusCode.INTERNAL,
                          "Error ensuring temporary dir: %s" % exc)

    def temp_dir_inner(self, storage_name, context, ensure=True):
        """Return the path to temporary directory for the given storage

        Similar to what Gitaly uses, with a dedicated path in order
        to be really sure not to overwrite anything. The important feature
        is that the temporary directory is under the root directory of
        the storage, hence on the same file system (atomic renames of
        other files from the storage, etc.)

        :param bool ensure: if ``True``, the temporary directory is created
           if it does not exist yet.
        :raises KeyError: if the storage is unknown
        :raises OSError: if creation fails.
        """
        temp_dir = os.path.join(self.storage_root_dir(storage_name, context),
                                b'+hgitaly', b'tmp')
        if not ensure:
            return temp_dir

        # not the proper time to switch everything to pathlib (operates on
        # str paths, but surrogates returned by os.fsdecode() seem to really
        # work well)
        to_create = []
        current = temp_dir

        while not os.path.exists(current):
            to_create.append(current)
            current = os.path.dirname(current)

        while to_create:
            # same mode as in Gitaly, hence we don't care about groups
            # although this does propagate the setgid bit
            os.mkdir(to_create.pop(), mode=0o755)

        return temp_dir

    def repo_workdirs_root(self, gl_repo: Repository, context):
        tmp = Path(os.fsdecode(
            self.temp_dir(gl_repo.storage_name, context, ensure=True)
        ))
        rpath = normalize_rpath(gl_repo)
        return tmp / 'workdirs' / rpath

    @contextmanager
    def working_dir(self, gl_repo: Repository, repo, context, changeset=None):
        """Provide a working directory updated to the given changeset.

        The working directory is part from the pool of reusable working
        directories and created if needed.
        """

        with working_directory(self.repo_workdirs_root(gl_repo, context),
                               repo,
                               changeset=changeset) as wd:
            # sadly, even passing repo.ui in share creation is not enough
            # for transient config items to be forwarded. Seems to be
            # unrelated to share-safe, since it is on by default since
            # Mercurial 6.1
            for section, key in CONFIG_TO_FORWARD_IN_WORKING_DIRS:
                wd.repo.ui.setconfig(section, key,
                                     repo.ui.config(section, key))
            yield wd

    def hg_init_repository(self, repository: Repository, context):
        """Initialize a mercurial repository from a request object.

        :return: ``None``: the resulting repository has to be loaded in the
           standard way, using its path.
        :raises RepositoryCreationError: and updates context with error
           code and details.
        """
        corr_logger = LoggerAdapter(logger, context)
        try:
            repo_path = self.repo_disk_path(repository, context)
        except KeyError:
            msg = "no such storage: %r" % repository.storage_name
            context.set_details(msg)
            corr_logger.error(msg)
            context.set_code(StatusCode.INVALID_ARGUMENT)
            raise RepositoryCreationError(repository)

        if os.path.lexists(repo_path):
            msg = ("creating repository: repository exists already")
            corr_logger.error(msg)
            context.set_details(msg)
            context.set_code(StatusCode.ALREADY_EXISTS)
            raise RepositoryCreationError(repository)

        try:
            corr_logger.info("Creating repository at %r", repo_path)
            hg.peer(self.ui, opts={}, path=repo_path, create=True)
        except OSError as exc:
            context.set_code(StatusCode.INTERNAL)
            context.set_details("hg_init_repository(%r): %r" % (repo_path,
                                                                exc))
            raise RepositoryCreationError(repository)

    def repo_command(self, repo, context, cmd, *args, **kwargs):
        """Call a Mercurial command on the given repository.

        :param str cmd: name of the command
        """
        cmd = cmdutil.findcmd(cmd.encode('ascii'), commands.table)[1][0]
        return cmd(repo.ui, repo, *args, **kwargs)

    def publish(self, changeset, context):
        self.repo_command(changeset.repo(), context, 'phase',
                          public=True,
                          draft=False,
                          secret=False,
                          force=False,
                          rev=[str(changeset.rev()).encode('ascii')])



# Copyright 2020 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from mercurial.node import (
    hex,
)

from heptapod.gitlab.tag import (
    gitlab_tag_ref,
)
from .gitlab_ref import (
    path_match_any,
)

EXCLUDED_TAG_TYPES = (b'local', None)  # 'tip' has type None


def iter_gitlab_tags(repo, deref=True):
    """Iterate on the tags that are visible to GitLab.

    Yield tag name and tag target

    :param bool deref: if ``True``, targets are :class:`changectx`, else
    hexadecimal node ids, as :class:`str`.
    """
    for name, node in repo.tags().items():
        if repo.tagtype(name) in EXCLUDED_TAG_TYPES:
            continue

        yield name, (repo[node] if deref else hex(node).decode('ascii'))


def iter_gitlab_tags_as_refs(repo, patterns=None, **kwargs):
    """Iterate on the repo tags, yielding full Git refs and change contexts.

    :param patterns: as in :func:`path_match_any` (presplit)
    """
    for name, target in iter_gitlab_tags(repo, **kwargs):
        ref_path = gitlab_tag_ref(name)
        if patterns and not path_match_any(ref_path, patterns):
            continue

        yield ref_path, target


def tagged_changeset(repo, tag_name):
    """Return the changeset that the given tags points to, or None.

    :return: a changectx.
    """
    tags = repo.tags()
    node = tags.get(tag_name)
    if node is None or repo.tagtype(tag_name) in EXCLUDED_TAG_TYPES:
        return None  # voluntarily explit

    return repo[node]



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest

from hgitaly.stub.server_pb2 import (
    ServerInfoRequest,
)
from hgitaly.stub.server_pb2_grpc import ServerServiceStub

from . import skip_comparison_tests

if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip


def test_compare_server_info(hgitaly_rhgitaly_comparison_no_repo):
    fixture = hgitaly_rhgitaly_comparison_no_repo
    rpc_helper = fixture.rpc_helper(
        stub_cls=ServerServiceStub,
        method_name='ServerInfo',
        request_cls=ServerInfoRequest,
        repository_arg=False,
    )

    rpc_helper.assert_compare()



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""High level utilities for Gitaly protocol messages."""

from mercurial.utils import stringutil as hg_stringutil
from mercurial import (
    node,
    pycompat,
)

from .stub.shared_pb2 import (
    Branch,
    CommitAuthor,
    CommitStatInfo,
    GitCommit,
    SignatureType,
    Tag,
)
from .stub.commit_pb2 import (
    ListLastCommitsForTreeResponse,
)
from .stub.mercurial_changeset_pb2 import (
    MercurialChangeset,
    MercurialChangesetField,
)
from google.protobuf import text_encoding
from google.protobuf.text_format import (
    _Printer as ProtobufPrinter,
    TextWriter,
)
from google.protobuf.timestamp_pb2 import Timestamp

from .diff import (
    diff_opts,
    chunk_stats,
)
from .gitlab_ref import reverse_refs


LONG_BYTES_TRUNCATION = 50
"""Threshold to truncate long bytes fields.

Keep in mind that with the escaping, a single byte can be represented as
up to 4 bytes.
"""


class Printer(ProtobufPrinter):
    """Modified version to truncate long bytes fields."""

    def PrintFieldValue(self, field, value):
        if isinstance(value, bytes) and len(value) > LONG_BYTES_TRUNCATION:
            self.out.write('"')
            self.out.write(text_encoding.CEscape(
                b'%s... (truncated, actual size %d bytes)' % (
                    value[LONG_BYTES_TRUNCATION:], len(value)),
                False))
            self.out.write('"')
        else:
            super(Printer, self).PrintFieldValue(field, value)


def message_to_string(message):
    """Using our own :class:`Printer` and our own defaults."""
    out = TextWriter(as_utf8=False)
    printer = Printer(
        out,
        indent=0,
        as_utf8=False,
        as_one_line=True,
        use_short_repeated_primitives=True,
        pointy_brackets=False,
        use_index_order=False,
        float_format=False,
        double_format=False,
        use_field_number=False,
        descriptor_pool=None,
        message_formatter=None,
        print_unknown_fields=False,
        force_colon=False)
    printer.PrintMessage(message)
    result = out.getvalue().rstrip()
    return result


class Logging:
    """Wrapper of requests and responses for sensible logging.

    Still has the formatting happen lazily, only if the message
    is emitted. We still pay the price of at least one instantiation though.
    """

    def __init__(self, message):
        self.msg = message

    def __repr__(self):
        return '%s { %s }' % (self.msg.__class__.__name__,
                              message_to_string(self.msg),
                              )

    __str__ = __repr__


def commit_author(ctx):
    """Produce a `GitCommit` message from a :class:`changectx` instance.
    """
    auth = ctx.user()
    date = Timestamp()
    # hg time resolution is the second, see
    # https://www.mercurial-scm.org/wiki/ChangeSet
    seconds, offset = ctx.date()
    date.FromSeconds(int(seconds))
    if offset <= 0:
        # Mercurial sign convention is opposite of UTC+2 etc.
        tz_sign = b'+'
        offset_abs = -offset
    else:
        tz_sign = b'-'
        offset_abs = offset

    # Mercurial offset is in seconds
    tz_hours = offset_abs // 3600
    tz_minutes = (offset_abs % 3600) // 60

    return CommitAuthor(
        email=hg_stringutil.email(auth),
        name=hg_stringutil.person(auth),
        date=date,
        timezone=b"%s%02d%02d" % (tz_sign, tz_hours, tz_minutes),
        )


def commit_stats(ctx):
    if len(ctx.parents()) > 1:
        # Gitaly does not provide the stats for merge commits
        # (see Gitaly Comparison tests)
        return None

    ctx_from = ctx.p1()
    adds = dels = files = 0
    chunks = ctx.diff(ctx_from, opts=diff_opts(ctx.repo()))
    for stats in chunk_stats(chunks, ctx_from, ctx):
        adds += stats.additions
        dels += stats.deletions
        files += 1
    return CommitStatInfo(additions=adds, deletions=dels, changed_files=files)


def ref_segments_match(ref, pattern):
    """Tell if given ref first segments are exactly as in pattern."""
    pat_split = pattern.rstrip(b'/').split(b'/')
    ref_split = ref.split(b'/', len(pat_split))
    for r, p in zip(ref_split, pat_split):
        if r != p:
            return False
    return True


def commit_referenced_by(ctx, patterns):
    """Return references to the given commit, as full ref paths."""
    if not patterns:
        return

    repo = ctx.repo()
    refs = reverse_refs(repo).get(ctx.hex(), ())
    return [r for r in refs
            if any(ref_segments_match(r, pat) for pat in patterns)]


def commit(ctx, with_short_stats=False, include_referenced_by=()):
    """Return :class:`GitCommit` object from Mercurial :class:`changectx`.

    :param bool with_short_stats: if ``True``, fill-in the ``short_stats``
       field of :class:`GitCommit` messages.

    subject and body are as in gitaly/internal/git/log/commitmessage.go::

      var body string
      if split := strings.SplitN(commitString, "\n\n", 2); len(split) == 2 {
          body = split[1]
      }
      subject := strings.TrimRight(strings.SplitN(body, "\n", 2)[0], "\r\n")

    See also slightly different stripping gitlab/lib/gitlab/git/commit.rb::

        message_split = raw_commit.message.split("\n", 2)
        Gitaly::GitCommit.new(
          id: raw_commit.oid,
          subject: message_split[0] ? message_split[0].chomp.b : "",
          body: raw_commit.message.b,
          parent_ids: raw_commit.parent_ids,
          author: gitaly_commit_author_from_rugged(raw_commit.author),
          committer: gitaly_commit_author_from_rugged(raw_commit.committer)
        )

    Special case for caller convenience::

        >>> commit(None) is None
        True
    """
    if ctx is None:
        return None

    descr = ctx.description()
    author = commit_author(ctx)
    short_stats = commit_stats(ctx) if with_short_stats else None
    referenced_by = commit_referenced_by(ctx, include_referenced_by)
    return GitCommit(id=ctx.hex(),
                     subject=descr.split(b'\n', 1)[0].rstrip(b'\r\n'),
                     body=descr,
                     body_size=len(descr),
                     parent_ids=[p.hex().decode()
                                 for p in ctx.parents()
                                 if p.rev() != node.nullrev],
                     author=author,
                     committer=author,
                     short_stats=short_stats,
                     referenced_by=referenced_by,
                     )


def branch(name, target):
    return Branch(name=name, target_commit=commit(target))


def tag(name, target, tagging=None, signature_type=None):
    """Produce a :class:`Tag` instance

    :param target: a :class:`changectx` for the target of the tag
    :param tagging: optional :class:`changectx` for the changeset that
                    sets the tag.
    :pram signature_type: a :class:`SignatureType` or ``None``.
    """
    if signature_type is None:
        signature_type = SignatureType.NONE

    if tagging is None:
        tag_id = message = tag_author = None
        message_size = 0
    else:
        # TODO SPEC comment in `shared.proto` says the message will be
        # nullified if above a certain size and the size will be carried over,
        # but it doesn't say whose responsibility it is to do that,
        # nor how that threshold is to be determined
        # (information should be found by reading Gitaly Golang source).
        message = tagging.description()
        message_size = len(message)
        tag_id = pycompat.sysstr(tagging.hex())
        tag_author = commit_author(tagging)

    return Tag(name=name,
               id=tag_id,
               target_commit=commit(target),
               message=message,
               message_size=message_size,
               tagger=tag_author,
               signature_type=signature_type
               )


CommitForTree = ListLastCommitsForTreeResponse.CommitForTree


def commit_for_tree(changeset, path):
    """Message indicating the last changeset having modified a path"""
    return CommitForTree(commit=commit(changeset), path_bytes=path)


def mercurial_changeset(changeset, fields=None) -> MercurialChangeset:
    """Serialize changeset information.

    :param changeset: a :class:`changectx` instance
    :param fields: if specified, only the given fields are set in the
      returned message, otherwise alll currently implemented are.
    """
    all_fields = fields is None
    attrs = dict(id=changeset.hex())
    if all_fields or MercurialChangesetField.PHASE in fields:
        attrs['phase'] = changeset.phase()
    if all_fields or MercurialChangesetField.OBSOLETE in fields:
        attrs['obsolete'] = changeset.obsolete()
    if all_fields or MercurialChangesetField.BRANCH in fields:
        attrs['branch'] = changeset.branch()
    if all_fields or MercurialChangesetField.TOPIC in fields:
        attrs['topic'] = changeset.topic()
    if all_fields or MercurialChangesetField.FQBN in fields:
        attrs['fqbn'] = changeset.fqbn()
    if all_fields or MercurialChangesetField.TOPIC_NAMESPACE in fields:
        attrs['topic_namespace'] = changeset.topic_namespace()
    if all_fields or MercurialChangesetField.PARENT_IDS in fields:
        attrs['parent_ids'] = [ctx.hex() for ctx in changeset.parents()]
    return MercurialChangeset(**attrs)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import re
import time

from heptapod.testhelpers import (
    LocalRepoWrapper,
)

from ..branch import (
    BranchSortBy,
    ensure_gitlab_branches_state_file,
    gitlab_branch_head,
    gitlab_branches_matcher,
    iter_gitlab_branches,
    iter_gitlab_branches_as_refs,
    iter_gitlab_branches_matching,
    sorted_gitlab_branches_as_refs,
)

from hgext3rd.heptapod.branch import (
    GITLAB_BRANCHES_MISSING,
    gitlab_branches,
    set_default_gitlab_branch,
    write_gitlab_branches,
)


def make_repo(path):
    return LocalRepoWrapper.init(path,
                                 config=dict(
                                     extensions=dict(topic='', evolve=''),
                                 ))


def test_iter_no_state_file(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    default_branch = b'branch/default'
    default_branch_ref = b'refs/heads/branch/default'

    base = wrapper.write_commit('foo')
    base_sha = base.hex().decode()

    assert list(iter_gitlab_branches(repo)) == [(default_branch, base)]
    assert list(iter_gitlab_branches(repo, deref=False)) == [
                (default_branch, base_sha)]

    assert list(iter_gitlab_branches_as_refs(repo)) == [
        (default_branch_ref, base)]
    assert list(iter_gitlab_branches_as_refs(repo, deref=False)
                ) == [(default_branch_ref, base_sha)]

    another = wrapper.write_commit('bar', branch='another',
                                   utc_timestamp=time.time() + 10)
    another_sha = another.hex().decode()
    another_branch_ref = b'refs/heads/branch/another'
    assert list(sorted_gitlab_branches_as_refs(
        repo, sort_by=BranchSortBy.FULL_REF_NAME
    )) == [
        (another_branch_ref, another),
        (default_branch_ref, base),
    ]
    assert list(sorted_gitlab_branches_as_refs(
        repo,
        sort_by=BranchSortBy.FULL_REF_NAME,
        deref=False,
    )) == [
        (another_branch_ref, another_sha),
        (default_branch_ref, base_sha),
    ]
    assert list(sorted_gitlab_branches_as_refs(
        repo,
        sort_by=BranchSortBy.UPDATED_ASC,
        deref=True,
    )) == [
        (default_branch_ref, base),
        (another_branch_ref, another),
    ]
    assert list(sorted_gitlab_branches_as_refs(
        repo,
        sort_by=BranchSortBy.UPDATED_ASC,
        deref=False,
    )) == [
        (default_branch_ref, base_sha),
        (another_branch_ref, another_sha),
    ]
    assert list(sorted_gitlab_branches_as_refs(
        repo,
        sort_by=BranchSortBy.UPDATED_DESC,
    )) == [
        (another_branch_ref, another),
        (default_branch_ref, base),
    ]

    assert list(sorted_gitlab_branches_as_refs(
        repo,
        sort_by=BranchSortBy.UPDATED_ASC,
        after=default_branch_ref,
    )) == [
        (another_branch_ref, another),
    ]


def test_named_branch_multiple_heads(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    default_branch = b'branch/default'

    # no head, no bookmark
    assert gitlab_branch_head(repo, default_branch) is None
    assert gitlab_branch_head(repo, b'zebook') is None

    # one head
    base = wrapper.write_commit('foo')
    assert gitlab_branch_head(repo, default_branch) == base
    assert list(iter_gitlab_branches(repo)) == [(default_branch, base)]

    # two heads, no bookmark
    head1 = wrapper.write_commit('foo')
    head2 = wrapper.write_commit('foo', parent=base)

    assert gitlab_branch_head(repo, default_branch) == head2
    assert set(iter_gitlab_branches(repo)) == {
        (default_branch, head2),
        (b'wild/' + head1.hex(), head1),
        (b'wild/' + head2.hex(), head2),
    }
    assert gitlab_branch_head(repo, b'wild/' + head1.hex()) == head1
    assert gitlab_branch_head(repo, b'wild/' + head2.hex()) == head2

    # one bookmarked head and one not bookmarked
    wrapper.command('bookmark', b'book2', rev=head2.hex())
    assert gitlab_branch_head(repo, default_branch) == head1
    assert set(iter_gitlab_branches(repo)) == {
        (default_branch, head1),
        (b'book2', head2),
    }
    assert gitlab_branch_head(repo, b'wild/' + head1.hex()) is None
    assert gitlab_branch_head(repo, b'wild/' + head2.hex()) is None
    assert gitlab_branch_head(repo, b'book2') == head2

    # all heads bookmarked
    wrapper.command('bookmark', b'book1', rev=head1.hex())
    assert gitlab_branch_head(repo, default_branch) is None
    assert set(iter_gitlab_branches(repo)) == {
        (b'book1', head1),
        (b'book2', head2),
    }

    # finally, a formally correct wild branch, with no corresponding changeset
    assert gitlab_branch_head(repo, b'wild/' + (b'cafe' * 10)) is None


def test_invalid_state_file_entry(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    ctx = wrapper.write_commit('foo')

    # invalid entry is just ignored, be it in favor of other type of ref
    # or within the same type
    write_gitlab_branches(repo,
                          {b'branch/default': ctx.hex(),
                           b'invalid': b'1234beef' * 5})
    assert list(iter_gitlab_branches(repo)) == [(b'branch/default', ctx)]


def test_ensure_state_file(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    wrapper.write_commit('foo')

    assert gitlab_branches(repo) is GITLAB_BRANCHES_MISSING
    ensure_gitlab_branches_state_file(repo)
    # the state file creation stil relying on Git refs, since we didn't
    # convert anything, it's normal for the resulting dict of branches to
    # be empty (a later version would perhaps recompute it from HGitaly).
    assert gitlab_branches(repo) is not GITLAB_BRANCHES_MISSING


def test_bookmarks_not_shadowing_default_branch(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    base = wrapper.write_commit('foo')  # not strictly necessary
    head1 = wrapper.write_commit('foo')

    default_branch = b'branch/default'
    set_default_gitlab_branch(repo, default_branch)

    wrapper.command('bookmark', b'book1', rev=head1.hex())
    assert gitlab_branch_head(repo, default_branch) == head1

    head2 = wrapper.write_commit('foo', parent=base)
    wrapper.command('bookmark', b'book2', rev=head2.hex())

    assert gitlab_branch_head(repo, default_branch) == head2
    assert set(iter_gitlab_branches(repo)) == {
        (b'book1', head1),
        (b'book2', head2),
        (default_branch, head2)
    }


def test_gitlab_branches_state_file(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    base = wrapper.commit_file('foo')
    default = wrapper.commit_file('foo')
    topic = wrapper.commit_file('foo', parent=base, topic='zztop')

    write_gitlab_branches(wrapper.repo,
                          {b'branch/default': default.hex(),
                           b'topic/default/zztop': topic.hex(),
                           })

    assert gitlab_branch_head(repo, b'branch/default') == default
    assert gitlab_branch_head(repo, b'branch/typo') is None

    assert dict(iter_gitlab_branches(repo)) == {
        b'branch/default': default,
        b'topic/default/zztop': topic,
    }

    assert dict(iter_gitlab_branches(repo, deref=False)) == {
        b'branch/default': default.hex().decode(),
        b'topic/default/zztop': topic.hex().decode(),
    }

    # for self coverage (will be also covered by service tests)
    assert dict(iter_gitlab_branches_matching(repo, [b'branch/*'])) == {
        b'branch/default': default,
    }

    # making one of the referenced changesets in the state file obsolete
    topic = wrapper.amend_file('foo')
    assert gitlab_branch_head(repo, b'topic/default/zztop') is None
    assert dict(iter_gitlab_branches(repo)) == {
        b'branch/default': default,
    }

    # two entries, one being invalid
    write_gitlab_branches(wrapper.repo,
                          {b'branch/default': b'0102cafe' * 5,
                           b'topic/default/zztop': topic.hex(),
                           })
    assert gitlab_branch_head(repo, b'branch/default') is None
    assert dict(iter_gitlab_branches(repo)) == {
        b'topic/default/zztop': topic,
    }


def test_gitlab_branches_matcher():
    for glob_patterns, rx_pattern in (
            ([b'branch/default'], b'^(branch/default)$'),
            # simple wildcard
            ([b'branch/*'], b'^(branch/.*)$'),
            # escaping of regexp metacharacters
            ([b'branch/def.ult'], br'^(branch/def\.ult)$'),
            # several patterns
            ([b'branch/*', b'topic/dev*/*'],
             br'^(branch/.*)|(topic/dev.*/.*)$'),
            # incoming patterns is an iterator with one element
            (iter([b'somebranch']), b'^(somebranch)$'),
            # incoming patterns is an iterator with several elements
            (iter([b'some', b'other']), b'^(some)|(other)$'),
    ):
        assert gitlab_branches_matcher(glob_patterns) == re.compile(rx_pattern)

    # special case with no pattern still returns a regular expression that
    # matches anything

    rx = gitlab_branches_matcher(())
    assert rx.search(b'anything') is not None
    assert rx.match(b'anything') is not None



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest

from mercurial import error

from heptapod.testhelpers import (
    LocalRepoWrapper,
)


def make_repo(path):
    return LocalRepoWrapper.init(path,
                                 config=dict(
                                     extensions=dict(hgitaly=''),
                                 ))


def test_description_revsets(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    wrapper.write_commit('foo', message="It does not appear in message")
    ctx2 = wrapper.write_commit('bar',
                                user="Jane Foo <jfoo@heptapod.test>",
                                message=r"nor in that one! Yes \ is antislash")
    ctx3 = wrapper.write_commit('baz', message="Finally, a foo in title")
    ctx4 = wrapper.write_commit('baz',
                                message="And also deeper in description\n\n"
                                "here is the foo that was wanted.")

    foo_revs = {ctx3.rev(), ctx4.rev()}

    # hpd_dsc_rx matches changeset description only
    assert set(repo.revs("hpd_dsc_rx('foo')")) == foo_revs
    # …and is case-sensitive
    assert len(repo.revs("hpd_dsc_rx('Foo')")) == 0
    # hpd_dsc_irx works the same except that it ignores case
    assert set(repo.revs("hpd_dsc_irx('Foo')")) == foo_revs
    # The pattern is really a regular expression
    assert set(repo.revs(r"hpd_dsc_rx('[fg]o{2,}\s')")) == foo_revs
    # special cases
    assert set(repo.revs(r"hpd_dsc_rx(r'\\')")) == {ctx2.rev()}

    with pytest.raises(error.ParseError) as exc_info:
        repo.revs("hpd_dsc_rx('[]')")
    assert 'invalid regexp' in str(exc_info.value)

    with pytest.raises(error.ParseError) as exc_info:
        repo.revs("hpd_dsc_rx()")
    assert 'requires a string' in str(exc_info.value)



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from grpc import (
    RpcError,
    StatusCode,
)
import pytest

from mercurial import (
    error as hg_error,
    phases,
)

from google.protobuf.timestamp_pb2 import Timestamp

from hgext3rd.heptapod.branch import (
    gitlab_branches,
)
from heptapod.testhelpers.git import GitRepo

from hgitaly.errors import (
    parse_assert_structured_error,
)
from hgitaly.stub.errors_pb2 import (
    ReferenceUpdateError,
)
from hgitaly.stub.mercurial_operations_pb2 import (
    CensorRequest,
    MergeBranchError,
    MergeBranchRequest,
    MercurialPermissions,
    PreCheckUpdateError,
    PublishChangesetRequest,
    MergeAnalysisRequest,
    MergeAnalysisResponse,
)
from hgitaly.stub.mercurial_operations_pb2_grpc import (
    MercurialOperationsServiceStub,
)
from hgitaly.tests.common import make_empty_repo_with_gitlab_state_maintainer
from .fixture import MutationServiceFixture

parametrize = pytest.mark.parametrize

WRITE_PERM = MercurialPermissions.WRITE
PUBLISH_PERM = MercurialPermissions.PUBLISH


class OperationsFixture(MutationServiceFixture):

    stub_cls = MercurialOperationsServiceStub

    def merge_analysis(self, **kw):
        return self.stub.MergeAnalysis(MergeAnalysisRequest(
            repository=self.grpc_repo, **kw))

    def publish(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        kw.setdefault('user', self.user)
        return self.stub.PublishChangeset(PublishChangesetRequest(**kw),
                                          metadata=self.grpc_metadata())

    def censor(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        kw.setdefault('user', self.user)
        return self.stub.Censor(CensorRequest(**kw),
                                metadata=self.grpc_metadata())

    def merge_branch(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        kw.setdefault('user', self.user)
        return self.stub.MergeBranch(MergeBranchRequest(**kw),
                                     metadata=self.grpc_metadata())


@pytest.fixture
def operations_fixture(grpc_channel, server_repos_root):
    with OperationsFixture(
            grpc_channel, server_repos_root,
            repo_factory=make_empty_repo_with_gitlab_state_maintainer,
    ) as fixture:
        yield fixture


def test_merge_analysis_ff(operations_fixture):
    repo_wrapper = operations_fixture.repo_wrapper
    commit_file = repo_wrapper.commit_file
    default_ctx = commit_file('foo')
    default_sha = default_ctx.hex().decode()
    topic_ctx = commit_file('foo', parent=default_ctx, topic='zetop')
    topic_first_hex = topic_ctx.hex()
    topic_first_sha = topic_first_hex.decode()

    merge_analysis = operations_fixture.merge_analysis
    assert (
        merge_analysis(source_revision=b'topic/default/zetop',
                       target_revision=b'branch/default')
        ==
        MergeAnalysisResponse(is_fast_forward=True,
                              source_node_id=topic_first_sha,
                              source_branch=b'default',
                              source_topic=b'zetop',
                              target_node_id=default_sha,
                              target_branch=b'default',
                              target_topic=b'',
                              target_is_public=False,
                              )
    )
    newtop_sha = commit_file('foo',
                             parent=topic_ctx,
                             topic='newtop').hex().decode()
    repo_wrapper.update(topic_ctx)
    repo_wrapper.amend_file('foo').hex()

    assert (
        merge_analysis(source_revision=b'topic/default/newtop',
                       target_revision=b'branch/default')
        ==
        MergeAnalysisResponse(is_fast_forward=True,
                              has_obsolete_changesets=True,
                              has_unstable_changesets=True,
                              source_node_id=newtop_sha,
                              source_branch=b'default',
                              source_topic=b'newtop',
                              target_node_id=default_sha,
                              target_branch=b'default',
                              target_topic=b'',
                              target_is_public=False,
                              )
    )
    assert (
        merge_analysis(source_revision=topic_first_hex,
                       target_revision=b'branch/default')
        ==
        MergeAnalysisResponse(is_fast_forward=True,
                              has_obsolete_changesets=True,
                              has_unstable_changesets=False,
                              source_node_id=topic_first_sha,
                              source_branch=b'default',
                              source_topic=b'zetop',
                              target_node_id=default_sha,
                              target_branch=b'default',
                              target_topic=b'',
                              target_is_public=False,
                              )
    )

    # error cases
    with pytest.raises(RpcError) as exc_info:
        merge_analysis(source_revision=b'unknown',
                       target_revision=b'branch/default')
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'source revision' in exc_info.value.details().lower()

    with pytest.raises(RpcError) as exc_info:
        merge_analysis(source_revision=b'topic/default/zetop',
                       target_revision=b'unknownn')
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'target revision' in exc_info.value.details().lower()


def test_merge_analysis_conflict(operations_fixture):
    repo_wrapper = operations_fixture.repo_wrapper
    commit_file = repo_wrapper.commit_file
    ctx0 = commit_file('foo')
    default_sha = commit_file('foo', content="default").hex().decode()
    repo_wrapper.set_phase('public', ['.'])  # also testing `target_is_public`
    topic_first_sha = commit_file('foo', parent=ctx0,
                                  topic='zetop', content="top"
                                  ).hex().decode()

    merge_analysis = operations_fixture.merge_analysis
    assert (
        merge_analysis(source_revision=b'topic/default/zetop',
                       target_revision=b'branch/default')
        ==
        MergeAnalysisResponse(is_fast_forward=False,
                              has_conflicts=True,
                              source_node_id=topic_first_sha,
                              source_branch=b'default',
                              source_topic=b'zetop',
                              target_node_id=default_sha,
                              target_branch=b'default',
                              target_topic=b'',
                              target_is_public=True,
                              )
    )

    # same without the conflicts check
    conflicts_check_skipped = merge_analysis(
        skip_conflicts_check=True,
        source_revision=b'topic/default/zetop',
        target_revision=b'branch/default',
    )
    assert conflicts_check_skipped.is_fast_forward is False
    assert conflicts_check_skipped.has_conflicts is False  # was really skipped

    # solving the conflict
    topic_fixed_sha = commit_file('foo', topic='zetop',
                                  content="default").hex().decode()
    assert (
        merge_analysis(source_revision=b'topic/default/zetop',
                       target_revision=b'branch/default')
        ==
        MergeAnalysisResponse(is_fast_forward=False,
                              has_conflicts=False,
                              source_node_id=topic_fixed_sha,
                              source_branch=b'default',
                              source_topic=b'zetop',
                              target_node_id=default_sha,
                              target_branch=b'default',
                              target_topic=b'',
                              target_is_public=True,
                              )
    )


@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_publish_changeset(operations_fixture, project_mode):
    fixture = operations_fixture
    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )

    wrapper = fixture.repo_wrapper

    ctx0 = wrapper.commit_file('foo')
    hex0 = ctx0.hex()
    default_head = wrapper.commit_file('foo')

    # watching GitLab branches move when a topical changeset gets
    # published is an important thing to test
    topical = wrapper.commit_file('foo', topic='zetop', message='topical')

    # because of the config set by fixture, operations through wrapper
    # are performed in all cases with Git mirroring.
    git_repo = GitRepo(str(wrapper.path) + '.git')
    initial_git_branches = git_repo.branches()

    default_ref = b'refs/heads/branch/default'
    top_ref = b'refs/heads/topic/default/zetop'

    # TODO make an obsolete and an unstable changeset, confirm it is
    # refused (or not, after all this is stricly for commit_files)

    # starting points
    assert ctx0.phase() == phases.draft
    assert fixture.list_refs() == {default_ref: default_head.hex().decode(),
                                   top_ref: topical.hex().decode()}

    fixture.publish(gitlab_revision=hex0, hg_perms=PUBLISH_PERM)
    wrapper.reload()
    assert wrapper.repo[hex0].phase() == phases.public

    # working with a symbolic revision
    fixture.publish(gitlab_revision=b'branch/default', hg_perms=PUBLISH_PERM)
    wrapper.reload()
    assert wrapper.repo[default_head.hex()].phase() == phases.public

    # publishing a topic
    fixture.publish(gitlab_revision=topical.hex(), hg_perms=PUBLISH_PERM)
    wrapper.reload()
    assert wrapper.repo[topical.hex()].phase() == phases.public

    assert fixture.list_refs() == {default_ref: topical.hex().decode()}

    if fixture.hg_native and not fixture.with_hg_git_mirroring:
        # expect Git branches not to have moved
        assert git_repo.branches() == initial_git_branches
    else:
        # expect Git branch to point on new commit
        assert git_repo.branches()[b'branch/default']['title'] == b'topical'

    # error cases
    with pytest.raises(RpcError) as exc_info:
        fixture.publish(gitlab_revision=b'not-resolvable',
                        hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'resolve revision' in exc_info.value.details().lower()

    with pytest.raises(RpcError) as exc_info:
        fixture.publish(hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'empty' in exc_info.value.details().lower()

    with pytest.raises(RpcError) as exc_info:
        fixture.publish(gitlab_revision=hex0, hg_perms=WRITE_PERM)


@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_merge_branch(operations_fixture, project_mode):
    fixture = operations_fixture
    merge = fixture.merge_branch

    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )
    wrapper = fixture.repo_wrapper

    gl_branch = b'branch/default'

    # because of the config set by fixture, this leads in all cases to
    # creation of a Git repo and its `branch/default` Git branch
    ctx0 = wrapper.commit_file('foo')
    initial_head = wrapper.commit_file('foo')
    ctx2 = wrapper.commit_file('foo', topic='zetop', message='fast-forward')
    sha2 = ctx2.hex().decode('ascii')
    conflict = wrapper.commit_file('foo', parent=ctx0, content='conflict',
                                   topic='conflicting')
    wrapper.commit_file('bar', branch='other', parent=ctx0)
    non_linear = wrapper.commit_file('old', parent=ctx0,
                                     topic='non-linear'
                                     ).hex().decode('ascii')

    # simple error cases
    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch, commit_id=sha2, message=b"whatever")
    assert exc_info.value.code() == StatusCode.PERMISSION_DENIED

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch, commit_id=sha2, message=b"whatever",
              hg_perms=WRITE_PERM)
    assert exc_info.value.code() == StatusCode.PERMISSION_DENIED

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch, commit_id=sha2, hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert exc_info.value.details() == 'empty message'

    with pytest.raises(RpcError) as exc_info:
        merge(message=b'whatever', commit_id=sha2, hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'branch name' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=b'topic/default/sometopic',
              message=b'whatever', commit_id=sha2, hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION
    assert 'named branches only' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id='12ca34fe' * 5,
              message=b'whatever',
              hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INTERNAL
    assert 'invalid commit' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id=sha2,
              expected_old_oid='12ca34fe' * 5,
              message=b'whatever',
              hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'cannot resolve' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id='not-a-hash',
              message=b'whatever',
              hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'parse commit' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id=sha2,
              expected_old_oid='not-a-hash',
              message=b'whatever',
              hg_perms=PUBLISH_PERM)
    assert exc_info.value.code() == StatusCode.INVALID_ARGUMENT
    assert 'parse commit' in exc_info.value.details()

    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id=non_linear,
              expected_old_oid=sha2,
              message=b'whatever',
              hg_perms=PUBLISH_PERM)
    _details, structured_err = parse_assert_structured_error(
        exc_info.value, MergeBranchError, StatusCode.FAILED_PRECONDITION)
    assert structured_err.reference_check == ReferenceUpdateError(
        reference_name=b'refs/heads/' + gl_branch,
        old_oid=sha2,
        new_oid=initial_head.hex().decode()
    )

    # semi-linear request but not fast-forwardable
    with pytest.raises(RpcError) as exc_info:
        merge(branch=gl_branch,
              commit_id=non_linear,
              message=b"whatever",
              hg_perms=PUBLISH_PERM,
              semi_linear=True)
    assert exc_info.value.code() == StatusCode.FAILED_PRECONDITION

    # successful fast-forward
    resp = merge(branch=gl_branch, commit_id=sha2,
                 hg_perms=PUBLISH_PERM,
                 message=b"whatever")
    wrapper.reload()

    assert resp.branch_update.commit_id == sha2
    assert wrapper.repo[ctx2.rev()].phase() == phases.public
    assert gitlab_branches(wrapper.repo)[gl_branch] == sha2.encode()

    if project_mode != 'native-project-without-git':
        git_repo = fixture.git_repo()
        assert git_repo.branch_titles()[gl_branch] == b'fast-forward'

    # successful actual merge
    resp = merge(branch=gl_branch,
                 commit_id=non_linear,
                 hg_perms=PUBLISH_PERM,
                 message=b"Actual merge!")
    wrapper.reload()

    merge_sha = resp.branch_update.commit_id
    assert merge_sha
    merge_changeset = wrapper.repo[merge_sha.encode()]
    assert merge_changeset.phase() == phases.public
    assert gitlab_branches(wrapper.repo)[gl_branch] == merge_changeset.hex()
    assert merge_changeset.p1().hex() == sha2.encode()
    assert merge_changeset.p2().hex() == non_linear.encode()
    assert merge_changeset.description() == b"Actual merge!"

    if project_mode != 'native-project-without-git':
        git_repo = fixture.git_repo()
        assert git_repo.branch_titles()[gl_branch] == b'Actual merge!'

    # conflict
    current_default_sha = merge_changeset.hex().decode()
    conflicting_sha = conflict.hex().decode()
    with pytest.raises(RpcError) as exc_info:
        resp = merge(branch=gl_branch,
                     commit_id=conflicting_sha,
                     hg_perms=PUBLISH_PERM,
                     message=b"whatever")
    details, merge_error = parse_assert_structured_error(
        exc_info.value, MergeBranchError, StatusCode.FAILED_PRECONDITION)
    conflict_error = merge_error.conflict
    assert conflict_error.conflicting_commit_ids == [current_default_sha,
                                                     conflicting_sha]
    assert conflict_error.conflicting_files == [b'foo']


@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_merge_branch_timestamp(operations_fixture, project_mode):
    fixture = operations_fixture
    merge = fixture.merge_branch

    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )
    wrapper = fixture.repo_wrapper

    gl_branch = b'branch/default'

    # because of the config set by fixture, this leads in all cases to
    # creation of a Git repo and its `branch/default` Git branch
    ctx0 = wrapper.commit_file('foo')
    ctx1 = wrapper.commit_file('foo')
    to_merge = wrapper.commit_file(
        'bar', branch='other', topic='zetop', parent=ctx0)

    solar_eclipse = 1624279076
    fixture.merge_branch(branch=gl_branch,
                         commit_id=to_merge.hex().decode(),
                         hg_perms=PUBLISH_PERM,
                         timestamp=Timestamp(seconds=solar_eclipse),
                         message=b'we have a date',
                         )
    wrapper.reload()
    merged = wrapper.repo[to_merge.rev()]
    assert merged.phase() == phases.public
    from mercurial import scmutil

    merge = scmutil.revsingle(wrapper.repo, b'default')
    assert merge.description() == b'we have a date'
    assert merge.date() == (solar_eclipse, 0)
    assert merge.p1().hex() == ctx1.hex()
    assert merge.p2() == merged

    if project_mode != 'native-project-without-git':
        git_repo = fixture.git_repo()
        git_branch = git_repo.branches()[gl_branch]
        assert git_branch['title'] == b'we have a date'
        assert git_repo.git(
            'log', '-n1', '--format=%as %cs', git_branch['sha']
        ).strip() == b'2021-06-21 2021-06-21'


def test_merge_branch_troubled_changesets(operations_fixture):
    fixture = operations_fixture
    wrapper = fixture.repo_wrapper

    gl_branch = b'branch/default'

    # stacked topics, and amending the first. The point of stacking
    # is that the amended changeset stays visible
    wrapper.commit_file('foo').hex().decode()
    amended = wrapper.commit_file('foo', topic='top1')
    wrapper.commit_file('foo', topic='top2')
    wrapper.update(amended)

    divergent = wrapper.amend_file('foo')  # not yet divergent

    def merge(source_changeset, **in_kw):
        out_kw = dict(branch=gl_branch,
                      hg_perms=PUBLISH_PERM,
                      commit_id=source_changeset.hex().decode(),
                      message=b'whatever')
        out_kw.update(in_kw)
        return fixture.merge_branch(**out_kw)

    with pytest.raises(RpcError) as exc_info:
        merge(amended)
    details, merge_error = parse_assert_structured_error(
        exc_info.value, MergeBranchError, StatusCode.FAILED_PRECONDITION)
    assert merge_error.pre_check == PreCheckUpdateError.OBSOLETE_CHANGESET

    # making a phase divergence
    wrapper.set_phase('public', [amended.hex()])

    with pytest.raises(RpcError) as exc_info:
        merge(divergent)
    details, merge_error = parse_assert_structured_error(
        exc_info.value, MergeBranchError, StatusCode.FAILED_PRECONDITION)
    assert merge_error.pre_check == PreCheckUpdateError.UNSTABLE_CHANGESET


@parametrize('project_mode', ('hg-git-project',
                              'native-project-without-git',
                              'native-project-with-git',
                              ))
def test_censor(operations_fixture, project_mode):
    fixture = operations_fixture
    # The `censor` extension is part of py-heptapod's `required.hgrc`, hence
    # always there in Heptapod context
    fixture.repo_wrapper.write_hgrc(dict(extensions=dict(censor='')))
    fixture.hg_native = project_mode != 'hg-git-project'
    if fixture.hg_native:
        fixture.with_hg_git_mirroring = (
            project_mode == 'native-project-with-git'
        )

    wrapper = fixture.repo_wrapper

    ctx0 = wrapper.commit_file('foo', content='oops, secret')
    hex0 = ctx0.hex()
    wrapper.commit_removal('foo')
    # censor cannot work if the file is in the working directory
    # TODO do in-method, we really want to be able to censor

    fixture.censor(changeset_node_id=hex0.decode(), file_path=b'foo')

    with pytest.raises(hg_error.Abort) as exc_info:
        wrapper.command('cat', b'/'.join((wrapper.repo.root, b'foo')),
                        rev=hex0)
    assert 'censored node' in str(exc_info.value)

    # error cases

    with pytest.raises(RpcError) as exc_info:
        fixture.censor(changeset_node_id='dead6789' * 5, file_path=b'foo')
    exc = exc_info.value
    assert exc.code() == StatusCode.INVALID_ARGUMENT
    assert 'changeset' in exc.details().lower()
    assert 'not found' in exc.details().lower()



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Fixture for Gitaly comparison tests based on Heptapod's Git mirroring."""
import attr
import contextlib
from copy import deepcopy
import functools
from itertools import (
    islice,
)
import logging
import pytest
import random
import warnings

import grpc

from mercurial_testhelpers.util import as_bytes

from hggit.git_handler import GitHandler
from heptapod.testhelpers.gitlab import GitLabMirrorFixture
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
)
from hgext3rd.heptapod.special_ref import (
    write_gitlab_special_ref,
    special_refs,
)
from hgitaly.errors import (
    STATUS_DETAILS_KEY,
    parse_structured_error,
)
from hgitaly import feature
from hgitaly.errors import HGITALY_ISSUES_URL
from hgitaly.gitlab_ref import (
    keep_around_ref_path,
    parse_keep_around_ref_path,
)
from hgitaly.servicer import SKIP_HOOKS_MD_KEY
from hgitaly.stub.shared_pb2 import Repository

try:
    from itertools import batched
except ImportError:  # Python<3.12
    def batched(it, n):
        # not sure in which version `while batch :=` is possible
        while True:
            batch = tuple(islice(it, n))
            if not batch:
                return
            yield batch

logger = logging.getLogger(__name__)


@attr.s
class GitalyComparison:
    hgitaly_channel = attr.ib()
    gitaly_channel = attr.ib()
    gitaly_repo = attr.ib()
    hgitaly_repo = attr.ib()
    gitlab_mirror = attr.ib()
    rhgitaly_channel = attr.ib(default=None)

    @property
    def hg_repo_wrapper(self):
        return self.gitlab_mirror.hg_repo_wrapper

    @property
    def git_repo(self):
        return self.gitlab_mirror.git_repo

    def rpc_helper(self, **kw):
        return RpcHelper(self, **kw)

    @functools.cached_property
    def hg_git(self):
        """The hg-git GitHandler instance, with SHA mapping preloaded.

        To invalidate this cached property, use :meth:`invalidate`
        """
        hg_repo = self.hg_repo_wrapper.repo
        hg_git = GitHandler(hg_repo, hg_repo.ui)
        hg_git.load_map()
        return hg_git

    def invalidate(self):
        """Invalidate all caches.

        In particular, reload the Mercurial repo
        """
        try:
            del self.hg_git
        except AttributeError:
            # the cached property has not been used yet, nothing to do
            pass

        self.hg_repo_wrapper.reload()
        self.gitlab_mirror.activate_mirror()

    def write_special_ref(self, ref_name, hg_sha):
        """Write a special ref in both repos

        :param ref_name: the special ref name (without `refs/`)
        :param bytes hg_sha: hexadecimal Mercurial Node ID of the changeset
          to point to

        :return: a pair of :class:`bytes` instance: (full ref path, git_sha)
        """
        git_sha = self.hg_git.map_git_get(hg_sha)
        if git_sha is None:
            raise LookupError("Git commit not found for %r" % hg_sha)

        hg_wrapper = self.hg_repo_wrapper
        git_repo = self.git_repo

        ref_path = b'refs/' + ref_name
        self.git_repo.write_ref(ref_path.decode(), git_sha)
        write_gitlab_special_ref(hg_wrapper.repo, ref_name, hg_sha)
        hg_wrapper.reload()
        assert ref_path in git_repo.all_refs()
        assert ref_name in special_refs(hg_wrapper.repo)
        return ref_path, git_sha

    def create_keep_around(self, hg_sha):
        """Create a keep-around ref in both repos

        :param bytes hg_sha: hexadecimal Mercurial Node ID of the changeset
          to point to

        On the Git side, the keep-around is set for the Git commit
        corresponding to the Mercurial commit.

        :return: a pair of :class:`bytes` instances:
          (full Mercurial ref path, full Git ref path)
        """
        git_sha = self.hg_git.map_git_get(hg_sha)
        hg_wrapper = self.hg_repo_wrapper
        git_repo = self.git_repo

        hg_ref_path = keep_around_ref_path(hg_sha)
        git_ref_path = keep_around_ref_path(git_sha)

        create_keep_around(hg_wrapper.repo, hg_sha)
        git_repo.write_ref(git_ref_path, git_sha)
        hg_wrapper.reload()
        return hg_ref_path, git_ref_path


@contextlib.contextmanager
def gitaly_comparison_fixture(server_repos_root,
                              gitaly_channel,
                              grpc_channel,
                              monkeypatch,
                              rhgitaly_channel=None,
                              ):
    common_relative_path = 'repo-' + hex(random.getrandbits(64))[2:]
    storage = 'default'

    gitaly_repo = Repository(relative_path=common_relative_path + '.git',
                             gl_repository='project-1234',
                             storage_name=storage)
    hgitaly_repo = Repository(relative_path=common_relative_path + '.hg',
                              storage_name=storage)

    hg_config = dict(phases=dict(publish=False),
                     ui=dict(username='Hgitaly Tests <hgitaly@heptapod.test>'),
                     extensions={name: '' for name in ('evolve',
                                                       'hggit',
                                                       'topic',
                                                       'hgitaly',
                                                       'heptapod')})
    with GitLabMirrorFixture.init(
            server_repos_root / storage,
            monkeypatch,
            common_repo_name=common_relative_path,
            hg_config=hg_config,
    ) as mirror:
        # configuration must be written in HGRC file, because
        # HGitaly server will load the repository independently.
        mirror.hg_repo_wrapper.write_hgrc(hg_config)
        mirror.activate_mirror()
        yield GitalyComparison(
            hgitaly_channel=grpc_channel,
            hgitaly_repo=hgitaly_repo,
            gitaly_channel=gitaly_channel,
            rhgitaly_channel=rhgitaly_channel,
            gitaly_repo=gitaly_repo,
            gitlab_mirror=mirror,
        )


class BaseRpcHelper:
    """Common helper for all comparisons.

    Will handle comparison between Gitaly and (HGitaly or RHGitaly) as well
    as comparison beween HGitaly and RHGitaly, or whatever comes next.
    """

    def __init__(self, comparison, stub_cls, method_name, request_cls,
                 repository_arg=True,
                 request_defaults=None,
                 feature_flags=(),
                 streaming_request_field=None,
                 streaming_request_chunk_size=1,
                 streaming=False):
        self.comparison = comparison
        self.stub_cls = stub_cls
        self.method_name = method_name
        self.request_cls = request_cls
        self.streaming = streaming
        self.streaming_request_field = streaming_request_field
        self.streaming_request_chunk_size = streaming_request_chunk_size
        self.repository_arg = repository_arg

        self.request_defaults = request_defaults
        self.streaming = streaming
        self.feature_flags = list(feature_flags)

        self.init_stubs()

    def init_stubs(self):
        """To be provided by subclasses."""
        raise NotImplementedError  # pragma no cover

    def grpc_metadata(self):
        mds = feature.as_grpc_metadata(self.feature_flags)
        mds.append((SKIP_HOOKS_MD_KEY, 'true'))
        return mds

    def stream_requests(self, **kwargs):
        to_stream = kwargs.pop(self.streaming_request_field, [])

        req = kwargs  # only for first request
        for chunk in batched(iter(to_stream),
                             self.streaming_request_chunk_size):
            req[self.streaming_request_field] = chunk
            yield self.request_cls(**req)
            req.clear()

    def rpc(self, backend, **kwargs):
        if self.repository_arg:
            kwargs.setdefault('repository', self.comparison.gitaly_repo)

        if self.streaming_request_field is not None:
            request = self.stream_requests(**kwargs)
        else:
            request = self.request_cls(**kwargs)

        meth = getattr(self.stubs[backend], self.method_name)
        metadata = self.grpc_metadata()
        if self.streaming:
            return [resp for resp in meth(request, metadata=metadata)]

        return meth(request, metadata=metadata)

    def apply_request_defaults(self, kwargs):
        defaults = self.request_defaults
        if defaults is not None:
            for k, v in defaults.items():
                kwargs.setdefault(k, v)

    def assert_compare_grpc_exceptions(self, exc0, exc1,
                                       same_details=True,
                                       vcses=('hg', 'git'),
                                       skip_structured_errors_issue=None,
                                       structured_errors_handler=None):
        """Comparison of exceptions once they are available.

        Singled out so that we can compare the errors provided by different
        gRPC methods, useful when we had to make a specific Mercurial method
        instead of just implementing some Gitaly method and still want to
        compare outcomes.
        """
        excs = [exc0, exc1]
        assert exc0.code() == exc1.code()
        if same_details:
            norm = self.error_details_normalizer
            details = [exc0.details(), exc1.details()]
            if norm is not None:  # pragma no cover
                for i, det in enumerate(details):
                    details[i] = norm(details[i], vcs=vcses[i])
            assert details[0] == details[1]

        # trailing metadata can bear a typed error gRPC message, which
        # is more important to compare than "details" (in the sense of
        # human-readable message), so let's insist on at least
        # having a tracking issue
        if skip_structured_errors_issue is None:
            # TODO check how to unskew that: RHGitaly currently adds
            # `content-length` and `date`, which Gitaly does not
            mds = [dict(exc.trailing_metadata()) for exc in (exc0, exc1)]
            for md in mds:
                md.pop('content-length', None)
                md.pop('date', None)
            if structured_errors_handler is not None:
                handler = structured_errors_handler
                # ignoring human-readable details, as they are much less
                # interesting and more volatile than the structured error
                # (and usually just repeat `exc.details()`)
                cls1 = handler[vcses[1] + '_cls']  # by default, git
                cls0 = handler.get(vcses[0] + '_cls', cls1)  # by default, hg
                handler_classes = [cls0, cls1]
                extracted = {}
                for i in range(2):
                    try:
                        code, _, error = parse_structured_error(
                            excs[i], handler_classes[i])
                        extracted[i] = dict(code=code, error=error)
                    except LookupError:  # pragma no cover
                        # unlikely case: caller is expecting structured errors
                        pass

                assert len(extracted) in (0, 2)  # None of them or both
                if extracted:
                    assert extracted[0]['code'] == extracted[1]['code']
                    if vcses[1] == 'git':
                        git_norm = handler.get('git_normalizer')
                        if git_norm is not None:
                            git_norm(extracted[1]['error'])
                        extracted[0]['error'] = handler['to_git'](
                            extracted[0]['error'])

                    assert extracted[0]['error'] == extracted[1]['error']

                # already done and a priori not directly comparable
                for md in mds:
                    md.pop(STATUS_DETAILS_KEY, None)
            assert mds[0] == mds[1]
        else:  # pragma no cover
            msg = ("Skipped comparison of structured errors, "
                   "until tracking issue %s/%d is done.") % (
                           HGITALY_ISSUES_URL, skip_structured_errors_issue)
            logger.warning(msg)
            warnings.warn(msg, RuntimeWarning)  # more visible in pytest output

        return exc0, exc1


class RpcHelper(BaseRpcHelper):
    """Encapsulates a comparison fixture with call and compare helpers.

    As Mercurial and Git responses are expected to differ (commit hashes and
    the like), this class provides a uniform mechanism to account for
    the expected difference, before finally asserting equality of
    the responses.

    # TODO much more to document.

    :attr:`feature_flags`: a mutable list of pairs such as
      ``(`my-flag`, True)``. The flags are sent to both servers.
    :attr:`response_sha_attrs`: Used to specify response attributes to
      convert to Git for comparison. See :meth:`attr_path_to_git` for
      specification.
    """

    def __init__(self, comparison, stub_cls, method_name, request_cls,
                 hg_server='hgitaly',
                 request_sha_attrs=(),
                 response_sha_attrs=(),
                 normalizer=None,
                 error_details_normalizer=None,
                 chunked_fields_remover=None,
                 chunks_concatenator=None,
                 **kwargs,
                 ):
        self.hg_server = hg_server
        super(RpcHelper, self).__init__(
            comparison, stub_cls, method_name, request_cls,
            **kwargs
        )
        self.request_sha_attrs = request_sha_attrs
        self.response_sha_attrs = response_sha_attrs
        self.normalizer = normalizer
        self.error_details_normalizer = error_details_normalizer
        self.chunked_fields_remover = chunked_fields_remover
        self.chunks_concatenator = chunks_concatenator

    def init_stubs(self):
        comparison, stub_cls = self.comparison, self.stub_cls

        if self.hg_server == 'rhgitaly':
            hg_channel = comparison.rhgitaly_channel
        else:
            hg_channel = comparison.hgitaly_channel
        self.stubs = dict(git=stub_cls(comparison.gitaly_channel),
                          hg=stub_cls(hg_channel))

    def hg2git(self, hg_sha):
        """Convert a Mercurial hex SHA to its counterpart SHA in Git repo.

        If not found in the Git Repo, the original SHA is returned, which
        is useful for tests about non existent commits.
        """
        # if hg_sha is None or not 40 bytes long it certainly won't
        # be found in the hg-git mapping, we don't need a special case
        # for that
        git_sha = self.comparison.hg_git.map_git_get(as_bytes(hg_sha))
        return hg_sha if git_sha is None else git_sha

    def normalize_keep_around(self, ref, vcs):
        if vcs != 'hg':
            return ref

        hg_sha = parse_keep_around_ref_path(ref)
        if hg_sha is None:
            # not a keep-around
            return ref
        return keep_around_ref_path(self.hg2git(hg_sha))

    def request_kwargs_to_git(self, hg_kwargs):
        git_kwargs = hg_kwargs.copy()
        for sha_attr in self.request_sha_attrs:
            sha = hg_kwargs.get(sha_attr)
            if sha is None:
                continue
            if isinstance(sha, list):
                git_kwargs[sha_attr] = [self.revspec_to_git(s) for s in sha]
            else:
                git_kwargs[sha_attr] = self.revspec_to_git(sha)
        return git_kwargs

    def revspec_to_git(self, revspec):
        """Convert revision specifications, including ranges to Git.

        This is to be improved as new cases arise.
        """
        is_bytes = isinstance(revspec, bytes)
        symdiff_sep = b'...' if is_bytes else '...'
        only_sep = b'..' if is_bytes else '..'

        for sep in (symdiff_sep, only_sep):
            if sep in revspec:
                # hg2git() defaulting rule will let symbolic revisions, such
                # as refs go through untouched
                return sep.join(self.hg2git(rev)
                                for rev in revspec.split(sep))
        # TODO implement caret, tilda etc.
        return self.hg2git(revspec)

    def response_to_git(self, resp):
        sha_attr_paths = [path.split('.') for path in self.response_sha_attrs]
        if self.streaming:
            for message in resp:
                self.message_to_git(message, sha_attr_paths)
        else:
            self.message_to_git(resp, sha_attr_paths)

    def message_to_git(self, message, attr_paths):
        for attr_path in attr_paths:
            self.attr_path_to_git(message, attr_path)

    def attr_path_to_git(self, message, attr_path):
        """Convert to Git part of message specified by an attr_path.

        :param attr_path: symbolic representation, as a succession of dotted
          attribute names. In case an attribute name ends with ``[]``, it
          is expected to be a simple list on which to iterate.
          Examples:
          - ``id``: convert ``message.id``
          - ``commits[].id``: for each element ``c`` of ``message.commits``,
            convert  ``c.id``
          - ``commits[].parent_ids[]`: for each element ``c`` of
            ``message.commits``, convert  all values in ``c.parent_ids``
        """
        obj = message
        trav = list(attr_path)
        while len(trav) > 1:
            attr_name, trav = trav[0], trav[1:]
            recurse = attr_name.endswith('[]')
            if recurse:
                attr_name = attr_name[:-2]
            # HasField cannot be used on repeated attributes, hence the elif
            elif not obj.HasField(attr_name):
                return
            obj = getattr(obj, attr_name)
            if recurse:
                for msg in obj:
                    self.message_to_git(msg, [trav])
                return

        obj_attr = trav[0]
        scalar_list = obj_attr.endswith('[]')
        if scalar_list:
            obj_attr = obj_attr[:-2]
        value = getattr(obj, obj_attr)

        if scalar_list:
            for i, sha in enumerate(value):
                value[i] = self.hg2git(sha)
        else:
            setattr(obj, obj_attr, self.hg2git(value))

    def call_backends(self, **hg_kwargs):
        """Call Gitaly and HGitaly with uniform request kwargs.

        To be used only if no error is expected.

        :param hg_kwargs: used as-is to construct the request for HGitaly,
          converted to Git and then to a request for Gitaly.
        """
        self.apply_request_defaults(hg_kwargs)

        git_kwargs = self.request_kwargs_to_git(hg_kwargs)

        git_response = self.rpc('git', **git_kwargs)
        hg_response = self.rpc('hg', **hg_kwargs)

        return hg_response, git_response

    def call_git_only(self, **hg_kwargs):
        """Call only Git, yet with Mercurial arguments.

        In some cases, comparison between Git and Mercurial is not directly
        possible (happens with mutations), but it is worthwile to check the
        assumptions on Gitaly behaviour that drive the HGitaly implementation.

        In that case, it is more convenient to take arguments as
        :meth:`assert_compare` does, hence providing conversion of arguments
        from Mercurial to Git.
        """
        self.apply_request_defaults(hg_kwargs)
        git_kwargs = self.request_kwargs_to_git(hg_kwargs)
        if self.repository_arg:
            git_kwargs.setdefault('repository', self.comparison.gitaly_repo)
        return self.rpc('git', **git_kwargs)

    def normalize_responses(self, hg_response, git_response):
        self.response_to_git(hg_response)
        norm = self.normalizer
        if norm is not None:
            norm(self, hg_response, vcs='hg')
            norm(self, git_response, vcs='git')

    def assert_compare(self, **hg_kwargs):
        hg_response, git_response = self.call_backends(**hg_kwargs)
        self.normalize_responses(hg_response, git_response)
        assert hg_response == git_response

    def assert_compare_aggregated(self,
                                  compare_first_chunks=True,
                                  check_both_chunked=True,
                                  **hg_kwargs):
        """Compare streaming responses with appropriate concatenation.

        Sometimes, it's unreasonable to expect HGitaly chunking to
        exactly match Gitaly's. This method allows to compare after
        regrouping the chunks, with the provided :attr:`chunks_concatenator`.

        Usually Gitaly returns small values within the first response only,
        to avoid the bandwidth waste of repetiting them. This helper
        checks that HGitaly does the same by comparing after applying
        :attr: `chunked_fields_remover` to as many responses as possible
        (typically the number of responses would differ).

        :param bool compare_first_chunk: if ``True``, the first chunks of
          both responses are directly compared (including main content). If
          ``False``, they are still compared, just ignoring main content.
        :param bool check_both_chunked: if ``True` checks that we get
          more than one response for both HGitaly and Gitaly
        :return: a pair: the first chunk of responses for Gitaly and HGitaly
          respectively, taken before normalization. This can be useful, e.g.,
          for pagination parameters.
        """
        assert self.streaming  # for consistency

        hg_resps, git_resps = self.call_backends(**hg_kwargs)
        original_first_chunks = deepcopy((git_resps[0], hg_resps[0]))

        self.normalize_responses(hg_resps, git_resps)
        if compare_first_chunks:
            assert hg_resps[0] == git_resps[0]

        if check_both_chunked:
            assert len(hg_resps) > 1
            assert len(git_resps) > 1

        concatenator = getattr(self, 'chunks_concatenator')
        fields_remover = getattr(self, 'chunked_fields_remover')
        assert concatenator(hg_resps) == concatenator(git_resps)

        for hg_resp, git_resp in zip(hg_resps, git_resps):
            if fields_remover is not None:
                fields_remover(hg_resp)
                fields_remover(git_resp)
            assert hg_resp == git_resp

        return original_first_chunks

    def assert_compare_errors(self, same_details=True,
                              skip_structured_errors_issue=None,
                              structured_errors_handler=None,
                              **hg_kwargs):
        """Compare errors returned by (R)HGitaly and Gitaly.

        :param:`structured_errors_handler`: if supplied, it is expected to be
          a :class:`dict` with the following keys:

          - `git_cls`: expected error class (gRPC message) from Gitaly
          - `hg_cls`: expected error class (gRPC message) from HGitaly
                      defaults to the value of `git_cls`
          - `to_git`: conversion callable from HGitaly's gRPC error message
            to Gitaly's
          - `git_normalizer`: additional normalizer to apply to Git error
            (some of them involve objects that cannot exist on the Mercurial
             side)
        """
        self.apply_request_defaults(hg_kwargs)

        git_kwargs = self.request_kwargs_to_git(hg_kwargs)
        with pytest.raises(grpc.RpcError) as exc_info_hg:
            self.rpc('hg', **hg_kwargs)
        with pytest.raises(grpc.RpcError) as exc_info_git:
            self.rpc('git', **git_kwargs)
        self.assert_compare_grpc_exceptions(
            exc_info_hg.value,
            exc_info_git.value,
            same_details=same_details,
            skip_structured_errors_issue=skip_structured_errors_issue,
            structured_errors_handler=structured_errors_handler,
        )


def normalize_commit_message(commit):
    """Remove expected differences between commits in Gitaly and HGitaly.

    Some are really testing artifacts, some have eventually to be removed.
    """
    # TODO tree_id should be replaced by HGitaly standard value
    # once HGitaly2 is the norm
    commit.tree_id = ''

    # hg-git may add a branch marker (this is just a test artifact)
    hg_marker = b'\n--HG--\n'
    split = commit.body.split(hg_marker, 1)
    if len(split) > 1:
        commit.body = split[0]
        commit.body_size = commit.body_size - len(split[1]) - len(hg_marker)

    # Either hg-git or Git itself adds a newline if there isn't one.
    # TODO investigate and if it is Git, add the newline in Mercurial
    # response.
    if not commit.body.endswith(b'\n'):
        commit.body = commit.body + b'\n'
        commit.body_size = commit.body_size + 1



# Copyright 2024 Georges Racinet <georges.racinet@cloudcrane.io>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging

from grpc import StatusCode

from ..errors import MercurialNotFound
from ..file_content import is_blob_generated
from ..servicer import HGitalyServicer

from ..stub.analysis_pb2 import (
    CheckBlobsGeneratedRequest,
    CheckBlobsGeneratedResponse,
)
from ..stub.analysis_pb2_grpc import AnalysisServiceServicer

base_logger = logging.getLogger(__name__)


class AnalysisServicer(AnalysisServiceServicer, HGitalyServicer):
    """AnalysisService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """

    def CheckBlobsGenerated(self, request: CheckBlobsGeneratedRequest,
                            context) -> CheckBlobsGeneratedResponse:
        first = True
        RespBlob = CheckBlobsGeneratedResponse.Blob
        for req in request:
            if first:
                # blobs are given by oid, hence as in direct changeset Node
                # IDs, the unfiltered repo is the right one for the task
                repo = self.load_repo(req.repository, context).unfiltered()
                first = False

            try:
                yield CheckBlobsGeneratedResponse(
                    blobs=(RespBlob(revision=blob.revision,
                                    generated=is_blob_generated(repo, blob))
                           for blob in req.blobs)
                )
            except (MercurialNotFound, ValueError):
                context.abort(StatusCode.INTERNAL,
                              "reading object: object not found")



# Copyright 2020-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import logging

from hgitaly import __version__
from ..servicer import HGitalyServicer
from ..stub.server_pb2 import (
    ServerInfoRequest,
    ServerInfoResponse,
)
from ..stub.server_pb2_grpc import ServerServiceServicer

base_logger = logging.getLogger(__name__)


class ServerServicer(ServerServiceServicer, HGitalyServicer):

    def ServerInfo(self,
                   request: ServerInfoRequest,
                   context) -> ServerInfoResponse:
        return ServerInfoResponse(server_version=__version__)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from datetime import (
    datetime,
    timedelta,
    timezone,
)
import grpc
import pytest
from mercurial import (
    pycompat,
    scmutil,
)
import time

from hgext3rd.heptapod.branch import (
    set_default_gitlab_branch,
)
from hgext3rd.heptapod.special_ref import (
    special_refs,
    write_gitlab_special_ref,
)
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
)
from hgitaly import feature
from hgitaly.errors import parse_assert_structured_error
from hgitaly.revision import (
    ZERO_SHA_STR,
)
from hgitaly.tests.common import (
    make_empty_repo,
    make_tree_shaped_repo,
)

from hgitaly.stub.shared_pb2 import (
    PaginationParameter,
    SortDirection,
)
from hgitaly.stub.ref_pb2 import (
    FindAllTagsRequest,
    RefExistsRequest,
    FindBranchRequest,
    FindLocalBranchesRequest,
    FindAllRemoteBranchesRequest,
    FindAllBranchesRequest,
    DeleteRefsRequest,
    ListBranchNamesContainingCommitRequest,
    ListRefsRequest,
    ListTagNamesContainingCommitRequest,
    GetTagMessagesRequest,
    FindRefsByOIDRequest,
    FindTagError,
    FindTagRequest,
)
from hgitaly.stub.ref_pb2_grpc import RefServiceStub

from .fixture import ServiceFixture


class RefFixture(ServiceFixture):

    stub_cls = RefServiceStub

    feature_flags = None

    def commit_file(self, *args, **kwargs):
        return self.repo_wrapper.commit_file(*args, **kwargs)

    def find_tag(self, **kw):
        return self.rpc('FindTag',
                        FindTagRequest(repository=self.grpc_repo, **kw))

    def find_all_tags(self, **kw):
        return self.stub.FindAllTags(
            FindAllTagsRequest(repository=self.grpc_repo, **kw))

    def find_local_branches(self, **kw):
        return self.stub.FindLocalBranches(
            FindLocalBranchesRequest(repository=self.grpc_repo, **kw),
            metadata=feature.as_grpc_metadata(self.feature_flags),
        )

    def list_refs(self, patterns=(b"refs/", ), **kw):
        return [(ref.name, ref.target)
                for resp in self.stub.ListRefs(
                        ListRefsRequest(repository=self.grpc_repo,
                                        patterns=patterns,
                                        **kw))
                for ref in resp.references]

    def find_refs_by_oid(self, **kw):
        kw.setdefault('repository', self.grpc_repo)
        return self.stub.FindRefsByOID(FindRefsByOIDRequest(**kw)).refs


@pytest.fixture
def ref_fixture(grpc_channel, server_repos_root):
    with RefFixture(grpc_channel, server_repos_root) as fixture:
        yield fixture


def test_tags(ref_fixture):
    wrapper = ref_fixture.repo_wrapper

    ctx = ref_fixture.commit_file('foo', message="The tagged chgs")
    wrapper.command('tag', b'v3.2.1', rev=ctx.hex())

    assert [n for n, _ in ref_fixture.list_refs(patterns=[b'refs/tags/'])
            ] == [b'refs/tags/v3.2.1']

    tag = ref_fixture.find_tag(tag_name=b'v3.2.1').tag
    assert tag.name == b'v3.2.1'
    target = tag.target_commit
    assert target.subject == b"The tagged chgs"

    assert [list(chunk.tags)
            for chunk in ref_fixture.find_all_tags()
            ] == [[tag]]

    # finally, structured error for a non existing tag
    with pytest.raises(grpc.RpcError) as exc_info:
        ref_fixture.find_tag(tag_name=b'does-not-exist')
    _details, error = parse_assert_structured_error(
        exc_info.value, FindTagError, grpc.StatusCode.NOT_FOUND)
    assert error.tag_not_found.reference_name == b'refs/tags/does-not-exist'


def test_find_branch(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    ctx = wrapper.write_commit('foo', message="Ze subject")
    resp = ref_stub.FindBranch(
        FindBranchRequest(repository=grpc_repo,
                          name=b'branch/default'))
    branch = resp.branch
    assert branch is not None
    assert branch.name == b'branch/default'
    assert branch.target_commit.id == ctx.hex().decode()

    resp = ref_stub.FindBranch(
        FindBranchRequest(repository=grpc_repo,
                          name=b'refs/heads/branch/default'))
    assert resp.branch == branch

    resp = ref_stub.FindBranch(
        FindBranchRequest(repository=grpc_repo,
                          name=b'cannot-be-found'))
    assert not resp.branch.name

    resp = ref_stub.FindBranch(
        FindBranchRequest(repository=grpc_repo,
                          name=b'refs/keeparound/012ca34fe56'))

    # There is no None in gRPC, just cascading default content (empty string).
    # We checked that Gitaly indeed uses the default `Branch(name=b'')`
    # to represent the absence of results.
    assert not resp.branch.name

    resp = ref_stub.FindAllBranches(
        FindAllBranchesRequest(repository=grpc_repo))
    branches = [br for chunk in resp for br in chunk.branches]
    assert len(branches) == 1
    assert branches[0].name == b'branch/default'
    assert branches[0].target == branch.target_commit

    resp = list(ref_stub.FindAllRemoteBranches(
        FindAllRemoteBranchesRequest(repository=grpc_repo)))
    assert not resp


def test_find_local_branches(ref_fixture):
    ctx = ref_fixture.commit_file('foo', message="Ze subject")

    ref_fixture.feature_flags = [
        ('simplify-find-local-branches-response', True),
    ]
    resp = ref_fixture.find_local_branches(
                pagination_params=PaginationParameter(limit=-1))
    branches = [br for chunk in resp for br in chunk.local_branches]
    assert len(branches) == 1
    assert branches[0].name == b'refs/heads/branch/default'
    target = branches[0].target_commit
    assert target.id == ctx.hex().decode()
    assert target.subject == b"Ze subject"


def test_find_local_branches_pagination(ref_fixture):
    # pagination doesn't preclude chunking inside each page
    for i in range(22):
        ref_fixture.commit_file('foo', branch='br%02d' % i, return_ctx=False)

    # limit=0 doesn't mean unlimited
    resp = list(ref_fixture.find_local_branches(
        pagination_params=PaginationParameter(limit=0)))
    assert len(resp) == 0

    # but no pagination_params does mean unlimited
    assert sum(len(resp.local_branches)
               for resp in ref_fixture.find_local_branches()) == 22

    resp = list(ref_fixture.find_local_branches(
        pagination_params=PaginationParameter(limit=21)))
    assert len(resp[0].local_branches) == 20
    # cursor aka page token is (not yet) returned
    assert [br.name for br in resp[1].local_branches] == [
        b'refs/heads/branch/br20']

    # but we can infer it
    resp = list(ref_fixture.find_local_branches(
        pagination_params=PaginationParameter(
            limit=6,
            page_token='refs/heads/branch/br20')))
    assert len(resp) == 1
    assert [br.name for br in resp[0].local_branches] == [
        b'refs/heads/branch/br21']


def test_find_local_branches_sort_by(ref_fixture):
    utc2 = timezone(timedelta(hours=2))

    ref_fixture.commit_file('foo', branch='aaa',
                            tz_datetime=(datetime(2021, 3, 1, 5, 57, 0,
                                                  tzinfo=utc2)  # 03:57 UTC
                                         ))
    ref_fixture.commit_file('foo', branch='bbb',
                            utc_datetime=(datetime(2021, 3, 1, 4, 0, 0)))
    ref_fixture.commit_file('foo', branch='ccc',
                            utc_datetime=(datetime(2021, 3, 1, 3, 0, 0)))

    SortBy = FindLocalBranchesRequest.SortBy
    pagination_params = PaginationParameter(limit=-1)

    def assert_branches_sorted_by(by, expected):
        assert [br.name
                for resp in ref_fixture.find_local_branches(
                        sort_by=by,
                        pagination_params=pagination_params
                )
                for br in resp.local_branches
                ] == [b'refs/heads/branch/' + ex for ex in expected]

    assert_branches_sorted_by(SortBy.NAME, [b'aaa', b'bbb', b'ccc'])
    assert_branches_sorted_by(SortBy.UPDATED_ASC, [b'ccc', b'aaa', b'bbb'])
    assert_branches_sorted_by(SortBy.UPDATED_DESC, [b'bbb', b'aaa', b'ccc'])


def test_ref_exists(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)

    def ref_exists(ref):
        return ref_stub.RefExists(RefExistsRequest(repository=grpc_repo,
                                                   ref=ref)).value

    with pytest.raises(grpc.RpcError) as exc_info:
        ref_exists(b'not-a-ref-path')
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    ctx = wrapper.write_commit('foo', message="Ze Foo")
    assert ref_exists(b'refs/heads/branch/default')
    assert not ref_exists(b'refs/heads/branch/other')
    assert not ref_exists(b'refs/heads/topic/default/zetop')
    assert not ref_exists(b'refs/tags/v3.2.1')

    wrapper.write_commit('zetop', topic='zetop')
    assert ref_exists(b'refs/heads/topic/default/zetop')

    wrapper.command('tag', b'v3.2.1', rev=ctx.hex())
    assert ref_exists(b'refs/tags/v3.2.1')
    assert not ref_exists(b'refs/tags/tip')

    # although we could resolve the hexadecimal node id from any
    # "wild" branch ref, it is just wrong to pretend it exists.
    assert not ref_exists(b'refs/heads/wild/' + ctx.hex())

    sref_name = b'pipelines/765'
    sref_path = b'refs/' + sref_name
    assert not ref_exists(sref_path)
    write_gitlab_special_ref(wrapper.repo, sref_name, ctx.hex())
    assert ref_exists(sref_path)

    keep_around = b'refs/keep-around/' + ctx.hex()
    assert not ref_exists(keep_around)
    create_keep_around(wrapper.repo, ctx.hex())
    assert ref_exists(keep_around)

    assert not ref_exists(b'refs/unknown/type/of/ref')


def test_delete_refs(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    repo = wrapper.repo

    def do_rpc(refs=(), except_prefixes=()):
        return ref_stub.DeleteRefs(
            DeleteRefsRequest(repository=grpc_repo,
                              refs=refs,
                              except_with_prefix=except_prefixes))

    with pytest.raises(grpc.RpcError) as exc_info:
        do_rpc(refs=[b'xy'], except_prefixes=[b'refs/heads'])
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    ctx = wrapper.write_commit('foo')
    wrapper.command('tag', b'v1.2.3')

    # Deleting a branch or a tag is forbidden
    for ref in (b'branch/default',
                b'refs/heads/branch/default',
                b'refs/tags/v1.2.3'):
        assert do_rpc([ref]).git_error

    special_ref_name = b'pipelines/256'
    special_ref_path = b'refs/' + special_ref_name

    write_gitlab_special_ref(repo, special_ref_name, ctx.hex())
    # double check
    assert special_refs(repo) == {special_ref_name: ctx.hex()}

    # go (it is normal for the client-side repo to need invalidation)
    assert not do_rpc([special_ref_path]).git_error
    # TODO use the future wrapper.reload()
    setattr(repo, '_gitlab_refs_special-refs', None)
    assert special_refs(repo) == {}

    # case where one mixes unknown and known refs
    write_gitlab_special_ref(repo, special_ref_name, ctx.hex())
    # TODO use the future wrapper.reload()
    setattr(repo, '_gitlab_refs_special-refs', None)
    # double check
    assert special_refs(repo) == {special_ref_name: ctx.hex()}
    mixed_refs = [special_ref_path, b'refs/merge-requests/12/head']
    assert not do_rpc(mixed_refs).git_error

    # TODO use the future wrapper.reload()
    setattr(repo, '_gitlab_refs_special-refs', None)
    assert special_refs(repo) == {}

    other_special_ref_name = b'environments/124'
    for prefix in (b'refs/environments',
                   b'refs/environ',
                   b'refs/environments/'):
        write_gitlab_special_ref(repo, special_ref_name, ctx.hex())
        write_gitlab_special_ref(repo, other_special_ref_name, ctx.hex())
        # double check
        assert special_refs(repo) == {special_ref_name: ctx.hex(),
                                      other_special_ref_name: ctx.hex()}

        assert not do_rpc(except_prefixes=[prefix]).git_error
        # TODO use the future wrapper.reload()
        setattr(repo, '_gitlab_refs_special-refs', None)
        assert special_refs(repo) == {other_special_ref_name: ctx.hex()}

    # exclusion matching both special refs
    write_gitlab_special_ref(repo, special_ref_name, ctx.hex())
    write_gitlab_special_ref(repo, other_special_ref_name, ctx.hex())
    assert not do_rpc(except_prefixes=[b'refs/environ',
                                       b'refs/pipel']).git_error
    assert special_refs(repo) == {special_ref_name: ctx.hex(),
                                  other_special_ref_name: ctx.hex()}


def test_list_branch_names_containing_commit(grpc_channel, server_repos_root):
    """Test ListBranchNamesContainingCommit on a repo a bit more spread
    """
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo, changesets = make_tree_shaped_repo(server_repos_root)

    def do_list(ctx, limit=0):
        chunks_iter = ref_stub.ListBranchNamesContainingCommit(
            ListBranchNamesContainingCommitRequest(
                repository=grpc_repo,
                commit_id=pycompat.sysstr(ctx.hex()),
                limit=limit,
            ))
        return [pycompat.sysstr(gl_branch) for chunk in chunks_iter
                for gl_branch in chunk.branch_names]

    wild1, wild2 = changesets['wild1'], changesets['wild2']
    top1, top2 = changesets['top1'], changesets['top2']

    wild_branch1 = 'wild/' + pycompat.sysstr(wild1.hex())
    wild_branch2 = 'wild/' + pycompat.sysstr(wild2.hex())
    assert all(do_list(ctx) == ['topic/default/zzetop']
               for ctx in [top1, top2])
    assert do_list(wild1) == [wild_branch1]
    assert set(do_list(wild2)) == {wild_branch2, 'branch/other'}
    assert set(do_list(changesets['other_base'])) == {'branch/other',
                                                      wild_branch1,
                                                      wild_branch2,
                                                      }
    for top in (top1, top2):
        assert do_list(top) == ['topic/default/zzetop']
    assert do_list(changesets['default']) == ['branch/default']
    all_branches = {'branch/other',
                    'branch/default',
                    'topic/default/zzetop',
                    wild_branch1,
                    wild_branch2,
                    }
    base = changesets['base']
    assert set(do_list(base)) == all_branches
    limited = set(do_list(base, limit=3))
    assert len(limited) == 3
    # until we have the ordering, we can only assert sub set.
    assert limited.issubset(all_branches)

    wrapper.update(top2)
    amended = wrapper.amend_file('foo')
    assert do_list(top2) == []
    assert do_list(amended) == ['topic/default/zzetop']


def test_list_tag_names_containing_commit(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    base = wrapper.write_commit('foo', message='Base')
    default = wrapper.write_commit('foo', message='Head of default')
    wrapper.command('tag', b'v3.2.1', rev=default.hex())

    other = wrapper.write_commit('foo', message='Start other',
                                 branch='other', parent=base)
    wrapper.command('tag', b'other-tag', rev=other.hex())

    def do_list(ctx, limit=0):
        chunks_iter = ref_stub.ListTagNamesContainingCommit(
            ListTagNamesContainingCommitRequest(
                repository=grpc_repo,
                commit_id=pycompat.sysstr(ctx.hex()),
                limit=limit,
            ))
        return [pycompat.sysstr(tag_name) for chunk in chunks_iter
                for tag_name in chunk.tag_names]

    all_tags = {'v3.2.1', 'other-tag'}
    assert set(do_list(base)) == all_tags
    assert do_list(default) == ['v3.2.1']
    assert do_list(other) == ['other-tag']

    limited = do_list(base, limit=1)
    assert len(limited) == 1
    assert limited[0] in all_tags

    wrapper.update(default)
    amended = wrapper.amend_file('foo')
    # The obsolete changeset is not filtered yet, as it has a child
    # that is not obsolete (but unstable).
    # TODO to avoid this nonsense in default opts, provide a
    # method of wrapper
    wrapper.command('evolve', stop=False, abort=False, rev=None,
                    dry_run=False, confirm=False, update=False,
                    phase_divergent=False, content_divergent=False,
                    orphan=True, tool=b'',
                    **{'continue': False, 'any': True, 'all': True,
                       'list': False})
    assert do_list(default) == []
    assert do_list(amended) == []


def test_get_tags_messages(grpc_channel, server_repos_root):
    ref_stub = RefServiceStub(grpc_channel)
    wrapper, grpc_repo = make_empty_repo(server_repos_root)
    wrapper.write_commit('foo', message='Base')
    wrapper.command('tag', b'v3.2.1', rev=b'.',
                    message=b"The tag message")
    tag_hex = wrapper.repo[b'tip'].hex()

    def do_list(tag_ids):
        resp_iter = ref_stub.GetTagMessages(
            GetTagMessagesRequest(
                repository=grpc_repo,
                tag_ids=(pycompat.sysstr(tag_id) for tag_id in tag_ids),
            ))
        return [pycompat.sysstr(resp.message) for resp in resp_iter]

    assert do_list([tag_hex]) == ["The tag message"]


def test_list_refs(ref_fixture):
    fixture = ref_fixture
    wrapper = fixture.repo_wrapper
    repo = wrapper.repo

    PSEUDO_REF_ALL = (b'ALL', ZERO_SHA_STR)

    # empty repo
    assert fixture.list_refs() == []
    assert fixture.list_refs(head=True) == []

    # with branches
    now = time.time()
    ctx1 = fixture.commit_file('foo', utc_timestamp=now - 30)
    sha1 = ctx1.hex().decode()
    ctx2 = fixture.commit_file('other', branch='aaa')
    sha2 = ctx2.hex().decode()

    assert fixture.list_refs(head=False) == [
        PSEUDO_REF_ALL,
        (b'refs/heads/branch/aaa', sha2),  # lexicographic sort ok
        (b'refs/heads/branch/default', sha1),
    ]
    # head (aka default GitLab branch)
    set_default_gitlab_branch(repo, b'branch/default')
    assert fixture.list_refs(head=True) == [
        PSEUDO_REF_ALL,
        (b'HEAD', sha1),
        (b'refs/heads/branch/aaa', sha2),
        (b'refs/heads/branch/default', sha1),
    ]

    # with tags
    wrapper.command('tag', b'v1.4', rev=ctx1.hex())
    tagging_sha = scmutil.revsingle(repo, b'.').hex().decode()
    assert fixture.list_refs() == [
        PSEUDO_REF_ALL,
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
        (b'refs/tags/v1.4', sha1),
    ]

    # special refs
    special = b'environments/17'
    special_ref = b'refs/' + special
    write_gitlab_special_ref(repo, special, ctx2.hex())
    assert fixture.list_refs() == [
        PSEUDO_REF_ALL,
        (special_ref, sha2),
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
        (b'refs/tags/v1.4', sha1),
    ]

    # with a keep around
    create_keep_around(repo, ctx1.hex())
    assert fixture.list_refs() == [
        PSEUDO_REF_ALL,
        (special_ref, sha2),
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
        (b'refs/keep-around/' + ctx1.hex(), sha1),
        (b'refs/tags/v1.4', sha1),
    ]

    # pattern used by gitaly-backup
    assert fixture.list_refs(patterns=[b'refs/']) == [
        PSEUDO_REF_ALL,
        (special_ref, sha2),
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
        (b'refs/keep-around/' + ctx1.hex(), sha1),
        (b'refs/tags/v1.4', sha1),
    ]

    # pattern used by Rails app for queries about branches and tags
    assert fixture.list_refs(patterns=[b'refs/heads/']) == [
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
    ]
    assert fixture.list_refs(patterns=[b'refs/tags/']) == [
        (b'refs/tags/v1.4', sha1),
    ]

    # more general patterns
    assert fixture.list_refs(patterns=[b'refs/heads/branch/a*']) == [
        (b'refs/heads/branch/aaa', tagging_sha),
    ]
    assert fixture.list_refs(patterns=[b'refs/t*/*']) == [
        (b'refs/tags/v1.4', sha1),
    ]
    assert fixture.list_refs(patterns=[b'refs/e*/*']) == [
        (special_ref, sha2),
    ]
    assert fixture.list_refs(patterns=[b'refs/heads/*/a*']) == [
        (b'refs/heads/branch/aaa', tagging_sha),
    ]
    # number of segments has to match
    assert fixture.list_refs(patterns=[b'*/a*']) == []
    assert fixture.list_refs(patterns=[b'refs/heads/br*']) == []
    assert fixture.list_refs(patterns=[b'refs/heads/br*/*']) == [
        (b'refs/heads/branch/aaa', tagging_sha),
        (b'refs/heads/branch/default', sha1),
    ]
    assert fixture.list_refs(
        patterns=[b'refs/keep-around/%s*' % ctx1.hex()[0:2]]
    ) == [
        (b'refs/keep-around/' + ctx1.hex(), sha1),
    ]

    # pointing_at_oids
    assert fixture.list_refs(pointing_at_oids=[ctx2.hex()]) == [
        (b'refs/environments/17', sha2)
    ]
    assert fixture.list_refs(head=True, pointing_at_oids=[ctx2.hex()]) == [
        (b'HEAD', sha1),
        (b'refs/environments/17', sha2)
    ]

    # sort options
    SortBy = ListRefsRequest.SortBy
    assert fixture.list_refs(patterns=[b'refs/heads/branch/'],
                             sort_by=SortBy(key=SortBy.Key.AUTHORDATE)) == [
                                 (b'refs/heads/branch/default', sha1),
                                 (b'refs/heads/branch/aaa', tagging_sha),
                             ]
    assert fixture.list_refs(patterns=[b'refs/heads/branch/'],
                             sort_by=SortBy(
                                 key=SortBy.Key.AUTHORDATE,
                                 direction=SortDirection.DESCENDING)) == [
                                 (b'refs/heads/branch/aaa', tagging_sha),
                                 (b'refs/heads/branch/default', sha1),
                             ]

    # reusing the fixture for the FindRefsByOID sibling method
    sha1 = ctx1.hex().decode()
    for oid in (sha1, sha1[:10]):
        assert fixture.find_refs_by_oid(oid=oid) == [
            'refs/heads/branch/default',
            'refs/tags/v1.4'
        ]
    assert fixture.find_refs_by_oid(oid=sha1, limit=1) == [
        'refs/heads/branch/default',
    ]
    assert fixture.find_refs_by_oid(oid=sha1, ref_patterns=['refs/tags/']) == [
        'refs/tags/v1.4',
    ]
    # TODO test 'creatordate' when implemented in ListRefs
    assert fixture.find_refs_by_oid(
        oid=sha1, sort_field='refname', limit=1
    ) == ['refs/heads/branch/default']
